# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from datetime import datetime

from chipscopy.api.noc.noc_perfmon_utils import NoCElement, get_noc_typedef_from_name, MAX_SAMPLES
from chipscopy.utils.logger import log

# %%
DOMAIN = "noc_perfmon"

pc_map = {"PC0": 0, "PC1": 1}

TG_EFF_SCALE = 1.042611


class HBMMC(NoCElement):
    pm_regs = [
        "hbmmc_mc_pm_read_pc0",
        "hbmmc_mc_pm_write_pc0",
        "hbmmc_mc_pm_read_ap_pc0",
        "hbmmc_mc_pm_write_ap_pc0",
        "hbmmc_mc_pm_sch_overhead_pc0",
        "hbmmc_mc_pm_sch_ooo_pc0",
        "hbmmc_mc_pm_sch_snub_pc0",
        "hbmmc_mc_pm_sch_rwswitch_pc0",
        "hbmmc_mc_pm_scrub_pc0",
        "hbmmc_mc_pm_rmw_pc0",
        "hbmmc_mc_pm_activate_pc0",
        "hbmmc_mc_pm_pchg_pc0",
        "hbmmc_mc_pm_preall_pc0",
        "hbmmc_mc_pm_sbrf_pc0",
        "hbmmc_mc_pm_grefresh_pc0",
        "hbmmc_mc_pm_read_pc1",
        "hbmmc_mc_pm_write_pc1",
        "hbmmc_mc_pm_read_ap_pc1",
        "hbmmc_mc_pm_write_ap_pc1",
        "hbmmc_mc_pm_sch_overhead_pc1",
        "hbmmc_mc_pm_sch_ooo_pc1",
        "hbmmc_mc_pm_sch_snub_pc1",
        "hbmmc_mc_pm_sch_rwswitch_pc1",
        "hbmmc_mc_pm_scrub_pc1",
        "hbmmc_mc_pm_rmw_pc1",
        "hbmmc_mc_pm_activate_pc1",
        "hbmmc_mc_pm_pchg_pc1",
        "hbmmc_mc_pm_preall_pc1",
        "hbmmc_mc_pm_sbrf_pc1",
        "hbmmc_mc_pm_grefresh_pc1",
        # TODO ??
        # 'hbmmc_mc_pm_pde_cnt',
        # 'hbmmc_mc_pm_sre_cnt',
        # 'hbmmc_mc_pm_pdx_cnt',
        # 'hbmmc_mc_pm_srx_cnt',
        # 'hbmmc_mc_pm_mrs_cnt',
        # 'hbmmc_mc_pm_interval_time'
    ]
    na_count = 2  # number of NoC Agent sub-blocks per HBMMC instance - AKA PSEUDO-CHANNEL
    port_count = 2  # number of psuedo channels per HBMMC instance
    pc_regexes = {
        "PC0": re.compile(".*pc0.*", re.IGNORECASE),
        "PC1": re.compile(".*pc1.*", re.IGNORECASE),
    }

    def __init__(
        self,
        name,
        alt_name,
        num_samples,
        sampling_period_ms,
        record_to_file,
        tslide=0,
        counter_freq_mhz=1000.0,
    ):
        """
        for HBM_MC 'HBM_MC_x' shall be the name and 'HBM_MC_NA_x' shall serve as the alternate name
        :param name:
        :param alt_name:
        """
        self.name = name
        self.alt_name = alt_name
        self.num_samples = num_samples
        self.sampling_period_ms = sampling_period_ms
        self.record_to_file = record_to_file
        self.node_type = get_noc_typedef_from_name(self.name)
        self.tslide = tslide
        self.counter_freq_mhz = counter_freq_mhz

        self.samples = {
            "read_bandwidth_pc0": [0] * MAX_SAMPLES,
            "read_bandwidth_pc1": [0] * MAX_SAMPLES,
            "write_bandwidth_pc0": [0] * MAX_SAMPLES,
            "write_bandwidth_pc1": [0] * MAX_SAMPLES,
            "total_bandwidth_pc0": [0] * MAX_SAMPLES,
            "total_bandwidth_pc1": [0] * MAX_SAMPLES,
            "ts": [],
            "type": self.node_type,
        }
        self.raw_trace_data = {}

        for metric in HBMMC.pm_regs:
            self.samples.update({f"{metric}": [0] * MAX_SAMPLES})

        self.axis_metrics = {
            "left_axis": {
                "bandwidth": [
                    "write_bandwidth_pc0",
                    "write_bandwidth_pc1",
                    "read_bandwidth_pc0",
                    "read_bandwidth_pc1",
                    "total_bandwidth_pc0",
                    "total_bandwidth_pc1",
                ],
                "latency": [],
                "mem": HBMMC.pm_regs,
            },
            "right_axis": {"bandwidth": [], "latency": [], "mem": []},
        }

    def get_axis_metrics(self, axis, view, pc):
        for metric in self.axis_metrics[axis][view]:
            if HBMMC.pc_regexes[pc].match(metric):
                yield metric

    def update_node(self, tcf_node, updated_keys):
        self.num_samples -= 1
        raw_trace_data = {
            "type": self.node_type,
            "name": self.name,
            # 'ts': (datetime.now() - self.start_time) + epoch,
            "ts": datetime.now(),
            "r": {},
            "w": {},
        }

        tcf_node_type = get_noc_typedef_from_name(tcf_node.Name)

        for reg in HBMMC.pm_regs:
            self.samples[f"{reg}"].append(tcf_node.props[f"{reg}"])

        # bandwidth
        rf = [0] * HBMMC.na_count
        read_bytes_per_s = [0] * HBMMC.na_count
        wf = [0] * HBMMC.na_count
        write_bytes_per_s = [0] * HBMMC.na_count

        for na_idx in range(
            0, HBMMC.na_count
        ):  # there are two noc agents per HBM MC and this is synonomous with the terminology pseudo-channel
            for port_idx in range(0, HBMMC.port_count):
                rf[na_idx] += tcf_node.props[
                    f"na{na_idx}.hbmmc_na{na_idx}_na_pm_read_data_p{port_idx}"
                ]
                wf[na_idx] += tcf_node.props[
                    f"na{na_idx}.hbmmc_na{na_idx}_na_pm_write_data_p{port_idx}"
                ]
        for pc_idx in range(0, HBMMC.na_count):
            # TODO remove wompus constant
            rf[pc_idx] = rf[pc_idx] * 128 / 8
            read_bytes_per_s[pc_idx] = rf[pc_idx] / (self.sampling_period_ms / 1000)
            self.samples[f"read_bandwidth_pc{pc_idx}"].append(read_bytes_per_s[pc_idx])
            # TODO remove wompus constant
            wf[pc_idx] = wf[pc_idx] * 128 / 8
            write_bytes_per_s[pc_idx] = wf[pc_idx] / (self.sampling_period_ms / 1000)
            self.samples[f"write_bandwidth_pc{pc_idx}"].append(write_bytes_per_s[pc_idx])
            self.samples[f"total_bandwidth_pc{pc_idx}"].append(
                read_bytes_per_s[pc_idx] + write_bytes_per_s[pc_idx]
            )

        # timestamp handling
        time_delta = raw_trace_data["ts"].strftime("%M:%S.") + str(
            int(raw_trace_data["ts"].microsecond / 1000)
        )

        rcmds = [tcf_node.props["hbmmc_mc_pm_read_pc0"], tcf_node.props["hbmmc_mc_pm_read_pc1"]]
        wcmds = [tcf_node.props["hbmmc_mc_pm_write_pc0"], tcf_node.props["hbmmc_mc_pm_write_pc1"]]
        # record bw
        log[DOMAIN].info(
            f"{self.name}: rb[pc0, pc1]: {rf}, wb[pc0, pc1]: {wf}, "
            + f"rbw[pc0, pc1]: {read_bytes_per_s}, wbw[pc0, pc1]: {write_bytes_per_s}, "
            + f"rcmd[pc0, pc1]: {rcmds}, wcmds[pc0, pc1]: {wcmds}"
        )
        self.samples["ts"].append(time_delta)
        self.trim_and_log(raw_trace_data)



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from pprint import pformat
from types import MethodType
from typing import ClassVar

from chipscopy.api import dataclass_fields, filter_props
from chipscopy.api.ila import ILA, ILAWaveform
from chipscopy.api.ila.ila_external_trace_data import (
    decode_external_data,
    EXTERNAL_TRACE_DATA_ALIGNMENT_BIT_COUNT,
)
from chipscopy.shared.ila_util import round_up_to_power_of_two, round_down_to_power_of_two
from chipscopy.shared.ila_data import (
    TraceInfo,
    calculate_trace_info_captured,
    calculate_raw_data_upload_byte_count,
    calculate_data_sample_byte_count,
)


@dataclass(frozen=True)
class ILAExternalTrace:
    """
    Trace Block related information, which get updated by calling one of these functions:

    - set_external_trace
    - upload_external_waveform
    - refresh_external_trace_status
    - run_basic_trigger
    - run_trigger_immediately

    """

    enabled: bool
    """
    
    - True - external trace is selected.
    - False - ILA core internal data buffer, is selected. Maximum data depth is 1024 samples.
    """
    address: int
    """Start address for waveform data in External Memory."""
    byte_size: int
    """Reserved waveform data byte size."""
    current_address: int
    """Trace Block current write address."""
    is_armed: bool
    """Trace Block armed status."""
    is_full: bool
    """Trace Block is_full status."""
    status_error_reserved: int
    """Undocumented."""
    current_stream_input_index: int = 0
    """Trace Block selected source index."""
    current_stream_input_name: str = ""
    """Trace Block selected source ILA cell name. Empty string means "Not reserved"."""
    _TRANSFER_OFFSET: ClassVar[int] = 64
    """Transfer unit adjust."""

    def __str__(self) -> str:
        p_dict = {
            name: value if isinstance(value, (bool, str)) else hex(value)
            for name, value in self.__dict__.items()
        }
        return pformat(p_dict, 2)

    def _arm(self, ila: ILA):
        ila.core_tcf_node.set_property({"deep_storage_enable": self.enabled})
        if not self.enabled:
            return

        if ila.control.window_size < 1024:
            raise ValueError(f'In external trace mode, "window_size" must be 1024 or larger.')

        if ila.control.window_count != 1:
            raise ValueError(
                f'Select "window_count=1. Only one window supported for external trace.'
            )

        needed_byte_count = calculate_external_trace_byte_count(
            ila, ila.control.window_size, ila.control.window_count
        )
        if needed_byte_count > self.byte_size:
            raise ValueError(
                f"Not enough external memory reserved. "
                f"Reduce windows size:{ila.control.window_size} or "
                f"window_count:{ila.control.window_count} or "
                f"increase external memory from 0x{self.byte_size:X} "
                f"to 0x{needed_byte_count:X} bytes."
            )

        ila.tcf_trace_block.acquire(
            ila.external_trace_stream_index,
            ila.name,
            self.address,
            needed_byte_count - ILAExternalTrace._TRANSFER_OFFSET,
        )
        refresh_external_trace_status(ila)


def _arm2(self):
    self.external_trace._arm(self)


def set_external_trace(ila: ILA, enable: bool, start_address: int, byte_count: int) -> None:
    """Trace Block setup. Applied during run trigger command.

    Args:
        ila (ILA): ILA object.
        enable (bool): True for external trace. False for ILA internal storage. ILA internal storage has room for 1024
        samples.
        start_address (int): Starting external memory address. start_address must be on a 0x1000 byte boundary.
        byte_count (int): Reserved trace data memory size.
    """

    def become_external_trace_ila():
        def raise_no_external_trace_ila():
            raise Exception(f'ILA "{ila.name}" does not have an associated Trace Block')

        # Monkey patch
        ila._arm2 = _arm2
        ila._arm2 = MethodType(_arm2, ila)

        if not ila.core_tcf_node["has_deep_storage"] and not ila._downstreams_refs:
            raise_no_external_trace_ila()
        trace_blocks = ila._device.trace_cores.filter_by(uuid=ila._downstreams_refs[0].uuid)
        if len(trace_blocks) == 0:
            raise_no_external_trace_ila()

        ila.tcf_trace_block = trace_blocks[0].core_tcf_node
        # Deduct 1, since downstream refs for Trace block in is in order "debug_hub", first ila, ...
        ila.external_trace_stream_index = ila._downstreams_refs[0].stream_index - 1

    if not hasattr(ila, "tcf_trace_block") or not hasattr(ila, "external_trace"):
        become_external_trace_ila()

    if enable:
        if byte_count <= 0:
            raise ValueError(
                f"External byte count 0x{byte_count:X} is too low to store any samples."
            )
        if start_address < 0:
            raise ValueError(f"Invalid external trace start_address 0x{start_address:X}.")

        if start_address % 0x1000 != 0:
            raise ValueError(
                f"Invalid external trace start_address 0x{start_address:X}. "
                f"Start address must to be on a 0x1000 byte boundary."
            )

    ila.external_trace = ILAExternalTrace(enable, start_address, byte_count, 0, False, False, 0)
    refresh_external_trace_status(ila)


def upload_external_waveform(ila: ILA, show_progress_bar: bool = True) -> bool:
    """
    Upload external waveform from DDR memory.

    Args:
        ila (ILA): ILA object.
        show_progress_bar (bool): Show progress bar while reading external trace memory.

    Returns: True, if waveform is captured and successfully uploaded.
    """

    def read_external_data(addr: int, byte_count: int) -> bytearray:
        buf = bytearray(byte_count)
        ila._device._dpc_read_bytes(addr, buf, show_progress_bar=show_progress_bar)
        return buf

    # Check ila dynamic status.
    if not ila.external_trace and not ila.external_trace.enabled:
        raise ValueError(f'External waveform is not enabled for ILA "{ila.name}."')

    ila.refresh_status()
    trace_info = TraceInfo(
        data_width=ila.static_info.data_width,
        trigger_position=ila.core_tcf_node["trigger_pos_readback"],
        window_count=ila.core_tcf_node["window_count_readback"],
        window_size=ila.core_tcf_node["window_depth_readback"],
    )
    trace_info = calculate_trace_info_captured(trace_info, ila.status.__dict__)
    if trace_info.window_count == 0:
        return False

    # Apparently some designs need to have "last_sample" register read, before "stopping"
    # the trace block (done by "tcf_trace_block.release")
    tb_last_sample = ila.tcf_trace_block.refresh_property_group(["last_sample"])

    # At least one window is available. Halt capture.
    ila.core_tcf_node.set_property({"arm_ila": False, "halt": True, "tas_enable": False})
    ila.core_tcf_node.commit_property_group(["__core_config", "control"])
    ila.tcf_trace_block.release(ila.external_trace_stream_index, ila.name, False)

    # upload and process data
    _, wave_data_byte_count = calculate_raw_data_upload_byte_count(
        trace_info, EXTERNAL_TRACE_DATA_ALIGNMENT_BIT_COUNT
    )
    data = read_external_data(ila.external_trace.address, wave_data_byte_count)

    trigger_info = ila.core_tcf_node.refresh_property_group(["trigger_info"])
    gap_at_end_sample_count = tb_last_sample["last_gap_count"]
    if trigger_info["last_gap_count_valid"]:
        gap_at_end_sample_count += trigger_info["last_gap_count"]
    # Only one window supported for now.
    beyond_last_record_adress = tb_last_sample["last_record_address"] - ila.external_trace.address
    trace, has_gaps = decode_external_data(
        data, trace_info, [beyond_last_record_adress], [gap_at_end_sample_count]
    )
    if trace_info.partial_window_sample_count:
        full_window_count = trace_info.window_count - 1
    else:
        full_window_count = trace_info.window_count

    # Update external memory status.
    refresh_external_trace_status(ila)

    # Create waveform instance
    wave_props = {
        "data": trace,
        "width": trace_info.data_width,
        "sample_count": trace_info.window_size * full_window_count
        + trace_info.partial_window_sample_count,
        "trigger_position": [trace_info.trigger_position] * trace_info.window_count,
        "window_size": trace_info.window_size,
        "probes": ila._make_waveform_probes(),
        "gap_index": trace_info.data_width if has_gaps else None,
    }

    ila.waveform = ILAWaveform(**wave_props)
    # In case trigger sample, was in a gap.
    if trigger_info["trigger_valid"]:
        ila.waveform.set_sample(trace_info.trigger_position, trigger_info["trigger_sample"])
    return True


def refresh_external_trace_status(ila: ILA) -> None:
    """Read Trace Block status. Update ILA.external_trace"""
    tb_status = ila.tcf_trace_block.refresh_property_group(["status"])
    et_props = ila.external_trace.__dict__
    et_props.update(tb_status)
    ila.external_trace = ILAExternalTrace(**filter_props(et_props, ILA_EXTERNAL_TRACE_MEMBERS))


def calculate_external_trace_byte_count(ila: ILA, sample_count: int, window_count: int) -> int:
    """
    Calculate byte count needed to store a certain number of samples.
    Argument "sample_count" is rounded up to nearest power-of-two value.

    Args:
        ila (ILA): ILA object.
        sample_count (int): Samples per window. Must be 1024 or larger.
        window_count (int): Number of windows.

    Returns: Number of bytes required to store samples for all windows.
    """
    # Round up sample_count to nearest power-of-two.
    one_window_sample_count = round_up_to_power_of_two(sample_count)
    # One bit for record type. Align sample on raw data words.
    one_sample_byte_count = calculate_data_sample_byte_count(
        EXTERNAL_TRACE_DATA_ALIGNMENT_BIT_COUNT, ila.static_info.data_width + 1
    )
    one_window_byte_count = one_window_sample_count * one_sample_byte_count
    return one_window_byte_count * window_count


def calculate_external_trace_one_window_sample_count(
    ila: ILA, byte_count: int, window_count: int
) -> int:
    """
    Calculate how many samples fit into one window.

    Args:
        ila (ILA): Ila object.
        byte_count (int): Total byte count for all windows.
        window_count (int): Number of windows.

    Returns: Max sample count, which fits into one single window. The result will be a power-of-two value.

    """
    # One bit overhead for record type.
    one_sample_byte_count = calculate_data_sample_byte_count(
        EXTERNAL_TRACE_DATA_ALIGNMENT_BIT_COUNT, ila.static_info.data_width + 1
    )
    one_window_byte_count = (byte_count // one_sample_byte_count) // window_count
    one_window_byte_count = round_down_to_power_of_two(one_window_byte_count)
    return one_window_byte_count


ILA_EXTERNAL_TRACE_MEMBERS = dataclass_fields(ILAExternalTrace)



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Dict, Any

from chipscopy.api.ibert.aliases import TX_RESET, TX_INJECT_ERROR
from chipscopy.api.ibert.serial_object_base import SerialObjectBase
from typing_extensions import final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.gt import GT  # noqa
    from chipscopy.api.ibert.link import Link
    from chipscopy.api.ibert.pll import PLL


@final
class TX(SerialObjectBase["GT", None]):
    def __init__(self, tx_info: Dict[str, Any], parent, tcf_node):
        SerialObjectBase.__init__(self, tx_info, parent, tcf_node)

        self.link: Optional[Link] = None
        """Link the TX is part of"""

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

    # def __repr__(self) -> str:
    #     return self.name

    @property
    def pll(self) -> PLL:
        """
        PLL driving the RX

        Returns:
            :py:class:`~chipscopy.api.ibert.pll.PLL` : PLL object
        """
        if len(self.parent.pll) == 1:
            return self.parent.pll[0]
        return self.parent.pll

    def reset(self):
        """
        Reset the TX
        """
        self.setup()

        if TX_RESET not in self.property_for_alias:
            raise RuntimeError(f"TX '{self.handle}' cannot be reset!")

        props = {self.property_for_alias[TX_RESET]: 0x1}
        self.property.set(**props)
        self.property.commit(self.property_for_alias[TX_RESET])

    def inject_error(self):
        """
        Inject error for the TX
        """
        self.setup()

        if TX_INJECT_ERROR not in self.property_for_alias:
            raise RuntimeError(f"TX '{self.handle}' cannot be reset!")

        props = {self.property_for_alias[TX_INJECT_ERROR]: 0x1}
        self.property.set(**props)
        self.property.commit(self.property_for_alias[TX_INJECT_ERROR])



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import dataclasses
import json
from collections import deque
from typing import Optional, Literal, List

from chipscopy.api.device.device_scanner import (
    ViewRecordType,
    JtagRecord,
    MemoryRecord,
    DebugcoreRecord,
    ChipscopeRecord,
)
from chipscopy.api.device.device_util import get_node_dna
from chipscopy.api.memory import Memory
from chipscopy.client import ServerInfo
from chipscopy.client.core import CoreParent
from chipscopy.client.debug_core_client import DebugCoreClientNode
from chipscopy.client.jtagdevice import JtagDevice, JtagCable
from chipscopy.client.mem import MemoryNode
from chipscopy.client.view_info import ViewInfo
from chipscopy.dm import Node


@dataclasses.dataclass
class DeviceSpec:
    """
    This support class keeps track of the low level TCF nodes for a device.
    Nodes can change - this class also tracks node events and marks invalid
    nodes as they change (ex: power cycle or program).

    What are we keeping track of? See below.

    ::

        jtag (ViewInfo)
        └── whole scan chain (JtagCable)          -> jtag_cable
            ├── arm_dap (JtagDevice)              -> jtag_dap
            └── xcvc1902 (JtagDevice)             -> jtag_device **

        memory (ViewInfo)
        ├── Versal xcvc1902 (MemoryNode)          -> memory_dap
               ...
        └── DPC (MemoryNode)                      -> memory_dpc

        debugcore (ViewInfo)
        ├── Versal xcvc1902 (Node)                -> debugcore_dap
        └── DPC xcvc1902 (DebugCoreClientNode)    -> debugcore_dpc

        chipscope (ViewInfo)
        ├── Versal xcvc1902 (Node)                -> chipscope_dap
        └── DPC xcvc1902 (CoreParent)             -> chipscope_dpc

    ** jtag_device is the default ctx that identifies the device when available
           (see get_device_ctx)
    """

    TargetType = Optional[Literal["DPC", "DAP"]]

    hw_server: ServerInfo
    cs_server: Optional[ServerInfo]
    jtag_cable_ctx: str = ""
    jtag_dap_ctx: str = ""
    jtag_device_ctx: str = ""
    jtag_index: Optional[int] = None
    memory_dap_ctx: str = ""
    memory_dpc_ctx: str = ""
    debugcore_dap_ctx: str = ""
    debugcore_dpc_ctx: str = ""
    chipscope_dpc_ctx: str = ""
    chipscope_dap_ctx: str = ""
    is_valid: bool = False
    device_ctx: str = ""
    device_view: str = ""

    @staticmethod
    def create_from_device_records(
        hw_server, cs_server, device_record_list: List[ViewRecordType]
    ) -> "DeviceSpec":
        """
        Factory to create a _DeviceNodes object correctly given the records list.
        """
        d = DeviceSpec(hw_server, cs_server)
        for rec in device_record_list:
            if type(rec) is JtagRecord:
                d.jtag_device_ctx = rec.ctx
                d.jtag_dap_ctx = rec.dap_ctx
                d.jtag_cable_ctx = rec.jtag_cable_ctx
                d.jtag_index = rec.jtag_index
            elif type(rec) is MemoryRecord:
                if rec.is_dap:
                    d.memory_dap_ctx = rec.ctx
                elif rec.is_dpc:
                    d.memory_dpc_ctx = rec.ctx
            elif type(rec) is DebugcoreRecord:
                if rec.is_dap:
                    d.debugcore_dap_ctx = rec.ctx
                elif rec.is_dpc:
                    d.debugcore_dpc_ctx = rec.ctx
            elif type(rec) is ChipscopeRecord:
                if rec.is_dap:
                    d.chipscope_dap_ctx = rec.ctx
                elif rec.is_dpc:
                    d.chipscope_dpc_ctx = rec.ctx

        d._pick_best_device_ctx()
        if d.device_ctx:
            d.is_valid = True
        return d

    def to_dict(self):
        fields = [field.name for field in dataclasses.fields(self)]
        retval = {}
        for f in fields:
            if f == "hw_server" or f == "cs_server":
                server = getattr(self, f)
                retval[f] = server.props.get("url", "")
            else:
                retval[f] = getattr(self, f)
        return retval

    def get_dna(self) -> Optional[int]:
        node = self.get_jtag_node()
        if node:
            return get_node_dna(node)
        else:
            node = self.get_debugcore_node(target=None, default_target="DAP")
            if node:
                return get_node_dna(node)

    def get_arch_name(self) -> str:
        node = self.get_jtag_node()
        arch_name = ""
        if node:
            arch_name = node.props.get("arch_name", "")
        else:
            node = self.get_debugcore_node(target=None, default_target="DAP")
            if node:
                try:
                    # In case of XVC cable, the jtag node does not exist. The
                    # dpc_driver_name is something we can use to identify xvc
                    dpc_driver_name = node.props.get("DpcDriverName", "")
                    if dpc_driver_name:
                        arch_name = "xvc"
                    else:
                        arch_name, _ = node.props.get("Name", "").split()
                except ValueError:
                    if node.props.get("Name", "").startswith("XVC"):
                        arch_name = "xvc"
                arch_name = arch_name.lower()
                if arch_name == "ultrascale":
                    arch_name = "uplus"
        return arch_name

    def get_part_name(self) -> str:
        node = self.get_jtag_node()
        part_name = ""
        if node:
            part_name = node.props.get("Name", "")
        else:
            node = self.get_debugcore_node(target=None, default_target="DAP")
            if node:
                try:
                    _, part_name = node.props.get("Name", "").split()
                except ValueError:
                    part_name = ""
        return part_name

    def to_json(self) -> str:
        raw_json = json.dumps(self.to_dict(), indent=4, default=lambda o: str(o))
        return raw_json

    def handle_node_update(self):
        # Check if device context changed - if so, mark the device invalid
        ...

    def get_device_node(self) -> Optional[Node]:
        if self.device_ctx:
            return self.get_node(self.device_view, self.device_ctx)
        else:
            return None

    def get_jtag_ctx(self) -> Optional[str]:
        return self.jtag_device_ctx

    def get_jtag_node(self) -> Optional[Node]:
        return self.get_node("jtag", self.jtag_device_ctx, JtagDevice)

    def get_jtag_cable_node(self) -> Optional[Node]:
        return self.get_node("jtag", self.jtag_cable_ctx, JtagCable)

    def get_jtag_cable_name(self) -> Optional[str]:
        node = self.get_jtag_cable_node()
        name = ""
        if node:
            name = node.props.get("Name", "")
        return name

    def get_memory_ctx(
        self, target: TargetType = None, default_target: TargetType = "DPC"
    ) -> Optional[str]:
        return self._pick_best_ctx(
            "memory", self.memory_dap_ctx, self.memory_dpc_ctx, target, default_target
        )

    def get_memory_node(
        self, target: TargetType = None, default_target: TargetType = "DPC"
    ) -> Optional[Memory]:
        ctx = self.get_memory_ctx(target=target, default_target=default_target)
        node = self.get_node("memory", ctx, MemoryNode)
        memory_view = self.hw_server.get_view("memory")
        try:
            memory = Memory.check_and_upgrade(memory_view, node)  # noqa
        except Exception:  # noqa
            memory = None
        return memory

    def get_all_memory_nodes(self) -> List[Memory]:
        all_nodes = []
        if self.memory_dpc_ctx:
            all_nodes.append(self.get_memory_node(target="DPC"))
        if self.memory_dap_ctx:
            memory_view = self.hw_server.get_view("memory")
            top = self.get_memory_node(target="DAP")
            dap_nodes = []
            nodes_to_travel = deque([top])
            while nodes_to_travel:
                node = nodes_to_travel.popleft()
                if MemoryNode.is_compatible(node):
                    memory_node = memory_view.get_node(node.ctx, MemoryNode)
                    memory_node = Memory.check_and_upgrade(memory_view, memory_node)
                    dap_nodes.append(memory_node)
                    nodes_to_travel.extendleft(memory_view.get_children(memory_node))
            all_nodes.extend(dap_nodes)
        return all_nodes

    def search_memory_node_deep(
        self, any_target: str = None, default_target: TargetType = "DPC"
    ) -> Optional[Memory]:
        # Go dumpster diving searching for the first match to the requested node
        # Note: here the any_target is the node name (not just DPC or DAP)...
        # This is an inefficient matching algorithm meant to maintain backward
        # compatibility with the previous memory search algorithm
        memory_view = self.hw_server.get_view("memory")
        if any_target == "DPC" or any_target == "DAP" or any_target is None:
            return self.get_memory_node(target=any_target, default_target=default_target)
        else:
            # TODO: Make a more efficient way to do this
            found_node = None
            top = self.get_memory_node(target="DAP", default_target=default_target)
            if top:
                nodes_to_travel = deque([top])
                memory_node_list = []
                while nodes_to_travel:
                    node = nodes_to_travel.popleft()
                    if node.props.get("Name") == any_target:
                        found_node = self.get_node("memory", node.ctx, MemoryNode)  # upgrade
                        break
                    if MemoryNode.is_compatible(node):
                        memory_node = memory_view.get_node(node.ctx, MemoryNode)
                        memory_node_list.append(memory_node)
                        nodes_to_travel.extendleft(memory_view.get_children(memory_node))
        if found_node:
            found_node = Memory.check_and_upgrade(memory_view, found_node)  # noqa
        return found_node

    def get_debugcore_ctx(
        self, target: TargetType = None, default_target: TargetType = "DPC"
    ) -> Optional[str]:
        return self._pick_best_ctx(
            "debugcore", self.debugcore_dap_ctx, self.debugcore_dpc_ctx, target, default_target
        )

    def get_debugcore_node(
        self, target: TargetType = None, default_target: TargetType = "DPC"
    ) -> Optional[Node]:
        ctx = self.get_debugcore_ctx(target=target, default_target=default_target)
        return self.get_node("debugcore", ctx, DebugCoreClientNode)

    def get_chipscope_ctx(
        self, target: TargetType = None, default_target: TargetType = "DPC"
    ) -> Optional[str]:
        return self._pick_best_ctx(
            "chipscope", self.chipscope_dap_ctx, self.chipscope_dpc_ctx, target, default_target
        )

    def get_chipscope_node(
        self, target: TargetType = None, default_target: TargetType = "DPC"
    ) -> Optional[Node]:
        ctx = self.get_chipscope_ctx(target=target, default_target=default_target)
        node = self.get_node("chipscope", ctx, CoreParent)
        return node

    def get_node(self, view_name: str, node_ctx: str, upgrade_cls=None) -> Optional[Node]:
        """Returns the node for a given context"""
        node = None
        try:
            view = self._get_view(view_name)
            if view:
                if upgrade_cls:
                    node = view.get_node(node_ctx, upgrade_cls)
                else:
                    node = view.get_node(node_ctx)
        except Exception:  # noqa
            node = None
        return node

    def _get_view(self, view_name: str) -> Optional[ViewInfo]:
        """reduce repetitive typing for getting the server/view"""
        if self.cs_server and view_name == "chipscope":
            return self.cs_server.get_view("chipscope")
        elif self.hw_server and view_name == "jtag":
            return self.hw_server.get_view("jtag")
        elif self.hw_server and view_name == "memory":
            return self.hw_server.get_view("memory")
        elif self.hw_server and view_name == "debugcore":
            return self.hw_server.get_view("debugcore")
        else:
            return None

    def _pick_best_device_ctx(self):
        """A device is associated with many contexts. We need to consistently
        choose one to represent the device. It should be the one least likely
        to be removed.

        If the device context changes due to some event, it invalidates the
        whole device.

        Set the best context to use
        """
        if self.jtag_device_ctx:
            self.device_ctx = self.jtag_device_ctx
            self.device_view = "jtag"
        elif self.debugcore_dpc_ctx:
            self.device_ctx = self.debugcore_dpc_ctx
            self.device_view = "debugcore"
        elif self.debugcore_dap_ctx:
            self.device_ctx = self.debugcore_dap_ctx
            self.device_view = "debugcore"

    def _pick_best_ctx(
        self,
        view_name: str,
        dap_ctx: str,
        dpc_ctx: str,
        target: TargetType,
        default_target: TargetType,
    ) -> str:
        """Helper - given a dap and dpc ctx and target, pick the best we have"""
        # TODO: determine if node invalid - don't assume context valid!!
        if target == "DAP":
            return dap_ctx  # explicitly specified
        elif target == "DPC":
            return dpc_ctx  # explicitly specified
        elif target is None:
            # Best guess based on availability and known default
            if view_name == "chipscope":
                view = self.cs_server.get_view(view_name)
            else:
                view = self.hw_server.get_view(view_name)
            try:
                is_dpc_valid = view.get_node(dpc_ctx) is not None
            except Exception:  # noqa
                is_dpc_valid = False
            try:
                is_dap_valid = view.get_node(dap_ctx) is not None
            except Exception:  # noqa
                is_dap_valid = False

            if default_target == "DPC" and is_dpc_valid:
                return dpc_ctx
            elif default_target == "DAP" and is_dap_valid:
                return dap_ctx
            else:
                # default target not available - return the best thing we have
                if is_dpc_valid:
                    return dpc_ctx
                elif is_dap_valid:
                    return dap_ctx
        return ""



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
from collections import UserList, defaultdict
from typing import Dict, List, Union, Iterable, Any, Optional, Set, NamedTuple

from antlr4 import ParserRuleContext
from antlr4.error.ErrorListener import ErrorListener
from antlr4.tree.Tree import TerminalNodeImpl

from chipscopy.api.ila.tsm.ILATsmParser import ILATsmParser
from chipscopy.api.ila.tsm.ILATsmVisitor import ILATsmVisitor
from chipscopy.api.ila.tsm.ila_tsm_data import (
    ILATsmData,
    ILATsmErrorListener,
    ILATsmMatchUnit,
    ILATsmCounterCompare,
    ILATsmAction,
    ILATsmActionPair,
    ILATsmBranch,
    ILATsmState,
    ILATsmCondition,
)
from chipscopy.api.ila import ILAProbe, ILA_MATCH_BIT_VALUES, ILA_MATCH_OPERATORS
from chipscopy.api.ila.tsm.ila_tsm_mapper import select_counter_actions
from chipscopy.shared.ila_util import hex_to_bin_str


class CounterMatch(NamedTuple):
    ctx: ILATsmParser.Counter_matchContext
    cmp: ILATsmCounterCompare


class ILATsmCheckAndBuild(ILATsmVisitor):
    # Semantic checks and build TSM back-end data structure.

    def __init__(
        self,
        tsm_data: ILATsmData,
        probe_enum_defs: Dict[str, enum.EnumMeta],
        state_names: List[str],
        counter_widths: List[int],
        error_listener: ErrorListener,
    ):
        self._tsm_data: ILATsmData = tsm_data
        self._probe_enum_defs: Dict[str, enum.EnumMeta] = probe_enum_defs
        self._counter_widths: List[int] = counter_widths
        self._error_listener: ILATsmErrorListener = error_listener
        self._state_names: List[str] = state_names
        self._counter_match_check: Dict[int, CounterMatch] = {}
        self._trigger_count: int = 0

    def aggregateResult(self, aggregate, next_result):
        """
        Aggregate results from non-terminal children, when calling self.visitChildren(ctx)
        Default overridden version just returns 'next_result' and does not aggregate.
        """
        if not next_result:
            return aggregate
        if not aggregate:
            return next_result
        if isinstance(aggregate, UserList):
            aggregate.append(next_result)
        else:
            aggregate = UserList([aggregate, next_result])
        return aggregate

    def _verify_counter_match(
        self, ctx: ILATsmParser.Counter_matchContext, cmp: ILATsmCounterCompare
    ):
        def ctx_to_str(ctx0: ILATsmParser.Counter_matchContext) -> str:
            return str(self.visit(ctx0.number()))

        prev: CounterMatch = self._counter_match_check.get(cmp.counter_index, None)
        if prev is None:
            # First found, add it.
            self._counter_match_check[cmp.counter_index] = CounterMatch(ctx, cmp)
            return

        # Compare the binary compare values.
        if prev.cmp.compare_value != cmp.compare_value:
            prev_ctx = prev.ctx.number()
            self._report_error(
                ctx.number(),
                "A counter must use the same compare value in the whole state machine.\n"
                f'The compare value "{ctx_to_str(ctx)}" conflicts with "{ctx_to_str(prev.ctx)}"\n'
                f"at line {prev_ctx.start.line}:{prev_ctx.start.column}.",
            )

    def _report_error(self, ctx: Union[ParserRuleContext, TerminalNodeImpl], msg: str):
        self._error_listener.report_error_ctx(ctx, msg)

    # Visit a parse tree produced by ILATsmParser#ila_tsm.
    def visitIla_tsm(self, ctx: ILATsmParser.Ila_tsmContext):
        tsm_states = self.visitChildren(ctx)
        if not isinstance(tsm_states, Iterable):
            tsm_states = [tsm_states]
        self._tsm_data.states = tsm_states
        select_counter_actions(self._tsm_data)

        if self._error_listener.get_error_count() == 0:
            verify_all_states_visited(tsm_states, self._error_listener)
        if self._trigger_count == 0:
            self._report_error(ctx, "The state machine need at least one trigger action.")

    def visitState_if(self, ctx: ILATsmParser.State_ifContext):
        state = ctx.IDENTIFIER().getText()
        state_index = self._state_names.index(state)
        branch_list = self.visit(ctx.if_condition())
        tsm_state = ILATsmState(state, state_index, branches=branch_list)
        return tsm_state

    def visitState_no_if(self, ctx: ILATsmParser.State_no_ifContext):
        state = ctx.IDENTIFIER().getText()
        state_index = self._state_names.index(state)
        action_block, next_state = self.visit(ctx.actions())
        tsm_branch = ILATsmBranch(
            index=0, actions=action_block, condition=None, next_state=next_state
        )
        tsm_state = ILATsmState(state, state_index, branches=[tsm_branch])
        return tsm_state

    # Visit a parse tree produced by ILATsmParser#if_condition.
    def visitIf_condition(self, ctx: ILATsmParser.If_conditionContext):
        """
        2 conditions, 3 action blocks, for if-then-elsif-then-else
        2 conditions, 2 action blocks, for if_then-elsif-then     (no else branch)
        1 condition,  2 action blocks, for if_then-else
        1 condition,  1 action block , for if_then                (no else branch)
        """

        full_conditions = [self.visit(cond) for cond in ctx.full_condition()]
        # action_blocks_and_next_states are a list of tuples:  [(ILATsmAction, goto_state)]
        action_blocks_and_next_states = [self.visit(action) for action in ctx.actions()]
        action_blocks, next_states = map(list, zip(*action_blocks_and_next_states))

        verify_all_if_condition_counters(ctx, full_conditions, self._error_listener)

        tsm_conditions = [ILATsmCondition.from_item_list(cond) for cond in full_conditions]
        if len(tsm_conditions) == len(action_blocks):
            self._report_error(ctx.actions(len(action_blocks) - 1), "Missing ELSE branch.")
        else:
            # Make all lists the same length.
            tsm_conditions.append(None)

        tsm_branches = [
            ILATsmBranch(index, actions, condition, next_state)
            for index, (actions, condition, next_state) in enumerate(
                zip(action_blocks, tsm_conditions, next_states)
            )
        ]

        return tsm_branches

    # Visit a parse tree produced by ILATsmParser#full_condition.
    def visitFull_condition(self, ctx: ILATsmParser.Full_conditionContext):
        condition_list = self.visitChildren(ctx)
        verify_condition(ctx, condition_list, self._error_listener)
        return condition_list

    # Visit a parse tree produced by ILATsmParser#condition_logic_op.
    def visitCondition_logic_op(self, ctx: ILATsmParser.Condition_logic_opContext):
        # Keep a flat list.
        condition_list = self.visit(ctx.condition(0))
        condition_list.append(ctx.LOGIC_OP().getText())
        condition_list.extend(self.visit(ctx.condition(1)))
        return condition_list

    # Visit a parse tree produced by ILATsmParser#condition_probe_match.
    def visitCondition_probe_match(self, ctx: ILATsmParser.Condition_probe_matchContext):
        return [self.visitChildren(ctx)]

    # Visit a parse tree produced by ILATsmParser#condition_counter_match.
    def visitCondition_counter_match(self, ctx: ILATsmParser.Condition_counter_matchContext):
        return [self.visitChildren(ctx)]

    def visitCondition_paren(self, ctx: ILATsmParser.Condition_parenContext):
        condition_list: list = self.visit(ctx.condition())
        has_counter = any(isinstance(item, ILATsmCounterCompare) for item in condition_list)
        has_probe = any(isinstance(item, ILATsmMatchUnit) for item in condition_list)
        has_parent_op = isinstance(ctx.parentCtx, ILATsmParser.Condition_logic_opContext)
        if has_counter and has_probe and has_parent_op:
            self._report_error(
                ctx,
                "Counter expression can not be together with a probe expression,"
                " inside nested parentheses.",
            )
        return condition_list

    # Visit a parse tree produced by ILATsmParser#probe_match.
    def visitProbe_match(self, ctx: ILATsmParser.Probe_matchContext):
        reverse_order = ctx.getChild(0) is not ctx.IDENTIFIER()
        number = self.visit(ctx.probe_number())
        probe_name = ctx.IDENTIFIER().getText()
        binary_number = "xxx"
        op = ctx.COMPARE_OP().getText()
        probe: ILAProbe = self._tsm_data.probes.get(probe_name, None)
        if not probe:
            self._report_error(ctx, f'Undefined probe: "{probe_name}"')
        else:
            enum_def = self._probe_enum_defs.get(probe_name, None)
            binary_number = verify_tsm_number(
                ctx, probe_name, probe.bit_width, op, number, self._error_listener, enum_def
            )

        return ILATsmMatchUnit(probe, op, binary_number, reverse_order)

    # Visit a parse tree produced by ILATsmParser#counter_match.
    def visitCounter_match(self, ctx: ILATsmParser.Counter_matchContext):
        number = self.visit(ctx.number())
        counter_name, counter_index, bit_width = self.get_counter_info_verify(ctx)
        op = ctx.COMPARE_OP().getText()
        if op not in ["==", "!="]:
            self._report_error(
                ctx,
                f'Operator "{op}" is not allowed in counter compare expressions. '
                'Only operators "==" and "!=" are supported for counters.',
            )

        if bit_width == 0:
            binary_number = "x"
        else:
            binary_number = verify_tsm_number(
                ctx, counter_name, bit_width, op, number, self._error_listener
            )
        cmp = ILATsmCounterCompare(counter_index, op, binary_number)
        self._verify_counter_match(ctx, cmp)
        return cmp

    # Visit a parse tree produced by ILATsmParser#actions.
    def visitActions(self, ctx: ILATsmParser.ActionsContext):
        actions = self.visitChildren(ctx)
        if not isinstance(actions, Iterable):
            actions = [actions]
        verify_action_block(ctx, actions, self._error_listener)
        next_state = 0
        for action in actions:
            if action.name == "goto":
                next_state = action.index
                break
        fsm_action = ILATsmAction.from_action_pair_list(actions)
        return fsm_action, next_state

    # Visit a parse tree produced by ILATsmParser#clear_flag.
    def visitAction_clear_flag(self, ctx: ILATsmParser.Action_clear_flagContext):
        return ILATsmActionPair(_lower(ctx.CLEAR_FLAG()), int(ctx.FLAG_NAME().getText()[-1]))

    # Visit a parse tree produced by ILATsmParser#set_flag.
    def visitAction_set_flag(self, ctx: ILATsmParser.Action_set_flagContext):
        return ILATsmActionPair(_lower(ctx.SET_FLAG()), int(ctx.FLAG_NAME().getText()[-1]))

    # Visit a parse tree produced by ILATsmParser#inc_counter.
    def visitAction_inc_counter(self, ctx: ILATsmParser.Action_inc_counterContext):
        _, counter_index, _2 = self.get_counter_info_verify(ctx)
        return ILATsmActionPair(_lower(ctx.INCREMENT_COUNTER()), counter_index)

    # Visit a parse tree produced by ILATsmParser#reset_counter.
    def visitAction_reset_counter(self, ctx: ILATsmParser.Action_reset_counterContext):
        _, counter_index, _2 = self.get_counter_info_verify(ctx)
        return ILATsmActionPair(_lower(ctx.RESET_COUNTER()), counter_index)

    # Visit a parse tree produced by ILATsmParser#goto.
    def visitAction_goto(self, ctx: ILATsmParser.Action_gotoContext):
        state_name = ctx.IDENTIFIER().getText()
        state_index = 0
        try:
            state_index = self._state_names.index(state_name)
        except ValueError:
            self._report_error(ctx.IDENTIFIER(), f'Unknown state "{state_name}".')

        return ILATsmActionPair(_lower(ctx.GOTO()), state_index)

    # Visit a parse tree produced by ILATsmParser#trigger.
    def visitAction_trigger(self, ctx: ILATsmParser.Action_triggerContext):
        self._trigger_count += 1
        return ILATsmActionPair(_lower(ctx.TRIGGER()), -1)

    # Visit a parse tree produced by ILATsmParser#enum_val.
    def visitEnum_val(self, ctx: ILATsmParser.Enum_valContext):
        return ctx.ENUM_VAL().getText()

    # Visit a parse tree produced by ILATsmParser#enum_string_val.
    def visitEnum_string_val(self, ctx: ILATsmParser.Enum_string_valContext):
        return ctx.ENUM_STRING_VAL().getText()

    # Visit a parse tree produced by ILATsmParser#int_val.
    def visitInt_val(self, ctx: ILATsmParser.Int_valContext):
        return ctx.INTEGER_VAL().getText()

    # Visit a parse tree produced by ILATsmParser#hex_val.
    def visitHex_val(self, ctx: ILATsmParser.Hex_valContext):
        return ctx.HEX_VAL().getText()

    # Visit a parse tree produced by ILATsmParser#bin_val.
    def visitBin_val(self, ctx: ILATsmParser.Bin_valContext):
        return ctx.BINARY_VAL().getText()

    def get_counter_info_verify(self, ctx) -> (str, int, int):
        counter_name = ctx.COUNTER_NAME().getText().upper()
        counter_index = int(counter_name[-1])
        bit_width = self._counter_widths[int(counter_index)]
        if bit_width == 0:
            self._report_error(ctx, f"{counter_name} is not supported by the ILA core.")
        return counter_name, counter_index, bit_width


def _lower(ctx: ParserRuleContext) -> str:
    return ctx.getText().lower()


def verify_all_states_visited(
    tsm_states: List[ILATsmState],
    error_listener: ILATsmErrorListener,
    visited_states: Set[int] = None,
    state_index: int = 0,
):
    if not tsm_states or state_index < 0 or state_index >= len(tsm_states):
        return
    if visited_states is None:
        visited_states = set()
    if state_index in visited_states:
        return

    visited_states.add(state_index)

    for branch in tsm_states[state_index].branches:
        if branch.next_state is not None:
            verify_all_states_visited(tsm_states, error_listener, visited_states, branch.next_state)

    if state_index == 0 and len(tsm_states) > len(visited_states):
        all_states = set(range(len(tsm_states)))
        unreached_states = all_states.difference(visited_states)
        unreached_state_names = ", ".join(
            tsm_states[index].name for index in sorted(unreached_states)
        )
        error_listener.report_error(
            None, None, f"The following states are not reachable: {unreached_state_names}."
        )


def verify_action_block(
    ctx: ILATsmParser.ActionsContext,
    actions: List[ILATsmActionPair],
    error_listener: ILATsmErrorListener,
):
    # Dict key is flag index or counter index.
    flags = defaultdict(list)
    counters = defaultdict(list)
    gotos = []
    trigger_count = 0

    for action in actions:
        if action.name == "trigger":
            trigger_count += 1
        elif action.name == "goto":
            gotos.append(action.index)
        elif action.name in ["set_flag", "clear_flag"]:
            flags[action.index].append(action.name)
        elif action.name in ["decrement_counter", "increment_counter", "reset_counter"]:
            counters[action.index].append(action.name)

    if len(gotos) > 1:
        error_listener.report_error_ctx(
            ctx, "Multiple GOTO statements are not allowed, in same the action block."
        )
    elif len(gotos) == 1 and actions[-1].name != "goto":
        error_listener.report_error_ctx(
            ctx, "The GOTO action must be the last action, in the action block."
        )
    elif len(gotos) + trigger_count != 1:
        error_listener.report_error_ctx(
            ctx, "An action block must have one GOTO action or one TRIGGER action."
        )

    if trigger_count > 1:
        error_listener.report_error_ctx(
            ctx, "Only one TRIGGER action is allowed in each action block."
        )

    for index, actions in flags.items():
        if len(actions) > 1:
            error_listener.report_error_ctx(
                ctx, f"$FLAG{index} may only have one action in each action block."
            )

    for index, actions in counters.items():
        if len(actions) > 1:
            error_listener.report_error_ctx(
                ctx, f"$COUNTER{index} may only have one action in each action block."
            )


def verify_condition(
    ctx: ILATsmParser.Full_conditionContext,
    condition_list: list,
    error_listener: ILATsmErrorListener,
):
    """
    Make sure
        - Only one counter, and the counter is first or last in the list.
        - Operators between probes are identical.
    Args:
        ctx (): parser tree node, has location used for error reporting.
        condition_list (): List made up of "||", "&&", ILATsmMatchUnit, ILATsmCounterCompare.
            Every 2nd item in the list is an operator.
        error_listener ():
    """
    counters = [isinstance(item, ILATsmCounterCompare) for item in condition_list]
    counter_count = counters.count(True)
    if counter_count > 1:
        error_listener.report_error_ctx(
            ctx, "Only one counter match comparison is allowed in a condition."
        )
    elif counter_count == 1 and len(condition_list) > 3 and not counters[0] and not counters[-1]:
        # if list is <= 3, the counter is in first or last position.
        error_listener.report_error_ctx(
            ctx, "Counter match comparison must be either first or last in a condition."
        )

    probe_ops = set()
    cl = condition_list
    for idx in range(1, len(cl), 2):
        if isinstance(cl[idx - 1], ILATsmMatchUnit) and isinstance(cl[idx + 1], ILATsmMatchUnit):
            probe_ops.add(cl[idx])

    if len(probe_ops) > 1:
        op_str = ", ".join(sorted(probe_ops))
        error_listener.report_error_ctx(
            ctx,
            "Different logic operator are not allowed between probe match comparisons,"
            " within an if-condition branch. "
            f"These operators are used: {op_str}.",
        )


def verify_all_if_condition_counters(
    ctx: ParserRuleContext, conditions: Union[list, Any], error_listener: ILATsmErrorListener
):
    # Verify only one counter is used in all conditions for one state.
    # in_conditions may be a list or a list of lists.
    counters_found = set()

    def find_counters(condition: Union[list, Any]):
        if isinstance(condition, ILATsmCounterCompare):
            counters_found.add(condition.counter_index)
        elif isinstance(condition, list):
            for item in condition:
                find_counters(item)

    find_counters(conditions)
    if len(counters_found) > 1:
        counters_str = ", ".join(f"$COUNTER{idx}" for idx in counters_found)
        error_listener.report_error_ctx(
            ctx,
            f'These counters "{counters_str}" are used in match expressions, in the same state. '
            "Only one counter may be used in the same state.",
        )


def verify_tsm_number(
    ctx: ParserRuleContext,
    name: str,
    bit_width: int,
    op: str,
    value: str,
    error_listener: ILATsmErrorListener,
    enum_def: Optional[enum.EnumMeta] = None,
) -> str:
    """Verify correctness of number and return binary value string."""

    number_error_count = 0

    def report_number(msg: str):
        nonlocal number_error_count
        error_listener.report_error_ctx(ctx, msg)
        number_error_count += 1

    def split_value_str(val: str) -> (int, str, str):
        q_pos = val.index("'")
        type_ch = val[q_pos + 1 : q_pos + 2].upper()
        width_prefix = int(val[:q_pos]) if q_pos > 0 else bit_width
        # Remove any quotes around enum value.
        number = val[q_pos + 2 :].strip('"')
        return width_prefix, type_ch, number

    def get_hex_char_bit_len(ch: str) -> int:
        if ch in "x01":
            return 1
        elif ch in "23":
            return 2
        elif ch in "4567":
            return 3
        else:
            return 4

    def verify_int_value(number: Union[str, int]) -> str:
        if isinstance(number, str):
            # Python string ints may use "_" between digits, but not in the first or last positions.
            int_value = int(number.strip("_"))
        else:
            int_value = number
        val_len = int.bit_length(int_value)
        if val_len > bit_width:
            report_number(
                f'"{name}" has bit width:{bit_width} and cannot be assigned'
                f' value "{number}" which has bit width:{val_len}.'
            )
            return None

        return f"{int_value:0{bit_width}b}"

    def verify_enum_value(number: str) -> str:
        # Only probes can have enum values. Grammar makes sure counter names are not using enum values.
        res = None
        if not enum_def:
            report_number(f'"Probe {name}" does not support enum values.')
        else:
            try:
                enum_int_val = enum_def[number].value
            except KeyError:
                report_number(f'Probe "{name}" does not have enum value "{number}" defined.')
            else:
                res = verify_int_value(enum_int_val)
        return res

    def verify_hex_value(hex_number: str):
        valid_hex_chars = "x0123456789abcdef_"
        value_no_underscore = "".join([ch for ch in hex_number.lower() if ch != "_"])
        hex_val_len = len(value_no_underscore)
        if hex_val_len == 0:
            report_number(f'"{name}" is assigned an invalid value "{hex_number}".')
            return None
        if not all([ch in valid_hex_chars for ch in value_no_underscore]):
            report_number(
                f'"{name}" value "{hex_number}" has invalid character(s).'
                f' Valid characters are {valid_hex_chars}".'
            )
            return None

        required_hex_ch_len = (bit_width + 3) // 4
        bin_val_len = (hex_val_len - 1) * 4 + get_hex_char_bit_len(value_no_underscore[0])
        if hex_val_len != required_hex_ch_len:
            report_number(
                f'"{name}" has bit width:{bit_width} requiring {required_hex_ch_len} hex character(s)'
                f' but value "{hex_number}" has {hex_val_len} hex character(s).'
            )
        elif bin_val_len > bit_width:
            # Too many bits in first hex char.
            report_number(
                f'"{name}" has bit width:{bit_width} and cannot be assigned'
                f' value "{hex_number}" which has bit width:{bin_val_len}.'
            )
        if (op not in ["==", "!="]) and ("x" in value_no_underscore):
            report_number(
                f'"{name}" is assigned value "{op} {hex_number}". '
                f"""The operator cannot be used when the value has a don't care "x"."""
            )
        return hex_to_bin_str(value_no_underscore, bit_width)

    def verify_bin_value(bin_number):
        # binary value
        bad_chars = set(bin_number) - set(ILA_MATCH_BIT_VALUES)
        if bad_chars:
            bad_char_str = "".join([ch for ch in bad_chars])
            report_number(
                f'"{name}" value "{bin_number}" has invalid character(s): "{bad_char_str}".'
                f' Valid characters are "{ILA_MATCH_BIT_VALUES}".'
            )
            return None

        value_no_underscore = "".join([ch for ch in bin_number.lower() if ch != "_"])
        val_len = len(value_no_underscore)
        if val_len != bit_width:
            report_number(
                f'"{name}" has bit width:{bit_width} and cannot be assigned'
                f' value "{bin_number}" which has bit width:{val_len}.'
            )
        if (op not in ["==", "!="]) and not all([ch in "01_" for ch in bin_number]):
            report_number(
                f'"{name}" is assigned value "{op} {bin_number}". '
                f'The operator "{op}" can only be used with "0" and "1" bit values.'
            )
        return value_no_underscore

    # Value has 3 parts <length prefix> <type char> <number>
    value_width, value_type_ch, value_number = split_value_str(value)

    if value_width <= 0 or value_width != bit_width:
        report_number(
            f'"{name}" is "{bit_width}" bits wide, but value "{value}" length prefix is "{value_width}".'
        )

    if op not in ILA_MATCH_OPERATORS:
        report_number(
            f'"{name}" uses operator "{op}" which is not one of the allowed operators: '
            f"{ILA_MATCH_OPERATORS}."
        )

    if bit_width == 1 and op not in ["==", "!="]:
        report_number(
            f'Operator "{op}" cannot be used for "{name}" which is 1 bit wide. '
            'Only operators "==" and "!=" are allowed.'
        )

    binary_result = None
    if number_error_count:
        binary_result = None
    elif value_type_ch == "U":
        binary_result = verify_int_value(value_number)
    elif value_type_ch == "E":
        binary_result = verify_enum_value(value_number)
    elif value_type_ch == "H":
        binary_result = verify_hex_value(value_number)
    elif value_type_ch == "B":
        binary_result = verify_bin_value(value_number)

    if number_error_count or binary_result is None:
        return "x" * bit_width
    else:
        return binary_result



# Copyright (C) 2024, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Dict, Union, Any

from chipscopy.api.ibert.aliases import CHILDREN, TYPE, PLL_SOURCE, PLL_KEY
from chipscopy.api.ibert.serial_object_base import SerialObjectBase
from chipscopy.api.ibert.pll import PLL
from more_itertools import one
from typing_extensions import final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.gt_group import GTGroup  # noqa


@final
class GTCOMMON(SerialObjectBase["GTGroup", None]):
    def __init__(self, gt_common_info: Dict[str, Any], parent, tcf_node):
        SerialObjectBase.__init__(self, gt_common_info, parent, tcf_node)

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

        self.pll: list[PLL] = []
        "PLL driving this GT_COMMON"

    def setup(self):
        if self.setup_done:
            return

        obj_info = self.core_tcf_node.get_obj_info(self.handle)

        self._build_aliases(obj_info)

        if not obj_info.get(CHILDREN):
            return

        # Build the child objects
        for child_name, child_obj_info in obj_info[CHILDREN].items():
            if child_obj_info[TYPE].startswith(PLL_KEY):
                obj = PLL(child_obj_info, self, self.core_tcf_node)
            else:
                continue

            obj.setup()
            self._children.append(obj)
            self.pll.append(obj)

        self.setup_done = True



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Union, Any

from chipscopy.api.containers import QueryList
from chipscopy.api.ibert.aliases import CHILDREN, GT_KEY, GT_COMMON_KEY, PLL_KEY, TYPE
from chipscopy.api.ibert.gt import GT
from chipscopy.api.ibert.gt_common import GTCOMMON
from chipscopy.api.ibert.pll import PLL
from chipscopy.api.ibert.serial_object_base import SerialObjectBase
from typing_extensions import final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.ibert import IBERT  # noqa


@final
class GTGroup(SerialObjectBase["IBERT", Union[GT, PLL]]):
    def __init__(self, gt_group_info: Dict[str, Any], parent, tcf_node):
        SerialObjectBase.__init__(self, gt_group_info, parent, tcf_node)

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

    # def __repr__(self) -> str:
    #     return self.name

    @property
    def gts(self) -> QueryList[GT]:
        """GTs in this GT Group"""
        self.setup()
        return self._children.filter_by(type=GT_KEY)

    @property
    def gtcommons(self) -> QueryList[GT]:
        """GTs in this GT Group"""
        self.setup()
        return self._children.filter_by(type=GT_COMMON_KEY)

    @property
    def plls(self) -> QueryList[PLL]:
        """PLL(s) in this GT Group"""
        self.setup()
        all_gt_group_plls = self._children.filter_by(type=PLL_KEY)
        for child in self.children:
            child.setup()
            all_gt_group_plls.extend(child.children.filter_by(type=PLL_KEY))
        if len(all_gt_group_plls) == 1:
            return all_gt_group_plls[0]
        return all_gt_group_plls

    def reset(self):
        """
        Reset the RXs, TXs and PLLs that are part of the GT Group
        """
        # Needn't call setup() here; Accessing 'gts' and 'plls' will trigger that call
        for gt in self.gts:
            gt.reset()
        for pll in self.plls:
            pll.reset()

    def setup(self):
        if self.setup_done:
            return

        self.core_tcf_node.setup_gt_group(self.name)
        obj_info = self.core_tcf_node.get_obj_info(self.handle, include_property="no")

        self._build_aliases(obj_info)

        if not obj_info.get(CHILDREN):
            return

        for child_info in obj_info[CHILDREN].values():
            obj: Union[GT, PLL]
            if child_info[TYPE] == GT_KEY:
                obj = GT(child_info, self, self.core_tcf_node)
            elif child_info[TYPE] == GT_COMMON_KEY:
                obj = GTCOMMON(child_info, self, self.core_tcf_node)
            elif child_info[TYPE].startswith(PLL_KEY):
                obj = PLL(child_info, self, self.core_tcf_node)
            else:
                continue

            obj.setup()
            self._children.append(obj)

        self.setup_done = True

        # Call this only after setup is complete.
        for gt in self.gts:
            gt.update_pll()



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
from typing import List, Dict, Callable

from chipscopy.api.ila.tsm.ila_tsm_data import (
    ILATsmData,
    ILATsmAction,
    MAX_STATE_COUNT,
    ILATsmState,
    ILATsmCondition,
    COUNTER_COUNT,
    ILATsmErrorListener,
    ILATsmMatchUnit,
    TABLE_ROWS_PER_STATE,
    MAX_CONDITION_COUNT_PER_STATE,
)
from chipscopy.shared.ila_mapper import IlaMapper, MapPortSeq, port_map_str_to_port_seqs, to_json


def select_counter_actions(tsm_data: ILATsmData) -> None:
    """
    Update branch actions, with counter selections, based their "next_state".
    """

    # Set state counter_index, by looking at branch conditions.
    for state in tsm_data.states:
        # Default value, when counter is not used in state conditions.
        counter_index = 0
        for branch in state.branches:
            if branch.condition and branch.condition.counter_compare:
                counter_index = branch.condition.counter_compare.counter_index
                break
        state.counter_index = counter_index

    # Update branch actions, with next counter to use.
    for state in tsm_data.states:
        for branch in state.branches:
            counter_index = tsm_data.states[branch.next_state].counter_index
            branch.actions = ILATsmAction.add_select_counter_action(branch.actions, counter_index)


def create_tsm_table_actions(tsm_data: ILATsmData) -> List[ILATsmAction]:
    """
    TSM table has 8 rows (or words) per state. There are maximum 16 states.
    TSM Address inputs, see class ILATsmInput.
    Args:
        tsm_data (ILATsmData): Parsed information.

    Returns: List of actions. One action object per "row" or address.

    """
    tsm_actions = []
    for state in tsm_data.states:
        # States have the 4 most significant bits in the TSM address.
        for if_mux in [False, True]:
            # "if mux" is the next bit in the TSM address
            for else_if_mux in [False, True]:
                # "else if mux" is the next bit in the TSM address
                for counter_mux in [False, True]:
                    # "counter mux" is the next bit in the TSM address, and the lsb.
                    action = get_actions_tsm_for_inputs(state, if_mux, else_if_mux, counter_mux)
                    tsm_actions.append(action)

    # Add un-used states
    unused_table_rows = (MAX_STATE_COUNT - len(tsm_data.states)) * TABLE_ROWS_PER_STATE
    un_used_state_actions = [ILATsmAction.DEFAULT] * unused_table_rows
    tsm_actions.extend(un_used_state_actions)
    return tsm_actions


def get_actions_tsm_for_inputs(
    state: ILATsmState, if_mux: bool, else_if_mux: bool, counter_mux
) -> ILATsmAction:
    """
    1 branch: No if-branch, no if-else branch exists.
    2 branches: An if-branch and an else-branch exist. No else-if branch.
    3 branches: All 3 branches exist: if, else-if, else:
    """
    branch = None
    if len(state.branches) > 1:
        # Try state if condition.
        if eval_condition(state.branches[0].condition, if_mux, counter_mux):
            branch = state.branches[0]

    if not branch and len(state.branches) > 2:
        # Try state "else-if" condition.
        if eval_condition(state.branches[1].condition, else_if_mux, counter_mux):
            branch = state.branches[1]

    if not branch:
        # Use last branch.
        branch = state.branches[-1]

    return branch.actions


def eval_condition(cond: ILATsmCondition, probe_mux: bool, in_counter_mux: bool) -> bool:
    """
    Evaluate a branch condition.

    Args:
        cond (ILATsmCondition): A branch condition.
        probe_mux (bool): True if the probe expression within the condition, is True.
        in_counter_mux (bool): True if the counter expression is True.

    Returns: True if the whole condition is True.

    """
    counter_mux = not in_counter_mux if cond.is_invert_counter_mux() else in_counter_mux
    if len(cond.match_units) == 0:
        # No probes in condition, just counter.
        res = counter_mux
    elif cond.counter_operator is None:
        # No counter, just probes in the condition.
        res = probe_mux
    elif cond.counter_operator == "||":
        # probe condition or'ed with counter.
        res = probe_mux or counter_mux
    else:
        # probe condition and'ed with counter.
        res = probe_mux and counter_mux
    return res


def map_tsm_actions_to_props(actions: [ILATsmAction]) -> Dict[str, str]:
    """

    Args:
        actions ([ILATsmAction): list of tsm actions

    Returns: TSM register value is binary string, with new-line char between tsm words.
    """
    bin_values = [f"{action.value:024b}" for action in actions]
    return {"fsm": "\n".join(bin_values)}


def map_tsm_counters_to_props(tsm_data: ILATsmData) -> Dict[str, str]:
    counter_values = {}
    for state in tsm_data.states:
        for branch in state.branches:
            if branch.condition and branch.condition.counter_compare:
                ccmp = branch.condition.counter_compare
                # Convert binary string to unsigned int string.
                counter_values[ccmp.counter_index] = str(int(ccmp.compare_value, 2))

    # Value is an integer, as a string.
    res = {f"counter{idx}_config": counter_values.get(idx, "0") for idx in range(COUNTER_COUNT)}
    return res


def map_tsm_mu_tc_to_props(
    tsm_data: ILATsmData, has_basic_capture_control: bool, error_listener: ILATsmErrorListener
) -> Dict[str, str]:
    """

    Args:
        tsm_data:
        has_basic_capture_control: True reserves first port match unit for basic capture control.
        error_listener:

    Returns: tcf ila property settings.

    """

    def create_mu_index_function(skip_first_mu: bool) -> Callable[[int, int], int]:
        """

        Args:
            skip_first_mu: True, means do not use the first match unit for each port.

        Returns: Function get_mu_index(port_index, mu_index_within_port) : ila_match_unit_index

        """
        # Create dictionary  (port_index, mu_index_within_port): mu_index
        mu_index_by_port_index_and_mu_index_with_port = {
            (pp.index, pmu_index): mu_index
            for pp in tsm_data.ports
            for pmu_index, mu_index in enumerate(pp.match_units)
        }

        def get_mu_index_fn(port_index: int, mu_index_within_port: int) -> int:
            if skip_first_mu:
                mu_index_within_port += 1
            mu_index = mu_index_by_port_index_and_mu_index_with_port.get(
                (port_index, mu_index_within_port), -1
            )
            if mu_index < 0:
                error_listener.report_error(
                    None, None, "Internal error mapping Trigger State Machine to Match Units."
                )
            return mu_index

        return get_mu_index_fn

    def add_probe_value_to_ports(
        ila_mapper: IlaMapper, tc_index: int, tsm_mu: ILATsmMatchUnit, seqs: [MapPortSeq]
    ) -> None:
        value = tsm_mu.compare_value
        idx = len(value)
        vals = []
        for seq in seqs:
            vals.append(value[idx - (seq.high_idx - seq.low_idx + 1) : idx])
            idx -= seq.high_idx - seq.low_idx + 1
        for seq, val in zip(seqs, vals):
            ila_mapper.add_probe_value(
                tc_index,
                tsm_mu.probe.name,
                seq.port_idx,
                seq.low_idx,
                tsm_mu.get_compare_operator(),
                val,
            )

    #
    mapper_ports = [(pp.bit_width, len(pp.match_units)) for pp in tsm_data.ports]
    if has_basic_capture_control:
        # First MU per port is reserved for capture control.
        mapper_ports = [
            (width, mu_count if mu_count <= 0 else mu_count - 1) for width, mu_count in mapper_ports
        ]
    mapper: IlaMapper = IlaMapper(is_basic_capture=False)
    mapper.add_ports(mapper_ports)

    for state in tsm_data.states:
        for branch_index, branch in enumerate(state.branches):
            if not branch.condition or not branch.condition.match_units:
                # No condition with probe match values, for this branch.
                continue

            condition_index = mapper.add_condition(
                branch.condition.get_mapper_probe_operator(),
                make_tc_name(state.index, branch_index),
            )
            for tsm_mu in branch.condition.match_units:
                seqs = port_map_str_to_port_seqs(tsm_mu.probe.name, tsm_mu.probe.map)
                add_probe_value_to_ports(mapper, condition_index, tsm_mu, seqs)

    # Map to match units, get result or overflow messages.
    status, overflow_message = mapper.map(skip_overflow_messages=False)
    if not status:
        error_listener.report_error(
            None, None, f"Unable to fit probe values on port match units:\n {overflow_message}"
        )
        # print(to_json(mapper))
        return {}
    else:
        # print(to_json(mapper))
        return mapper.get_reg_mu_mapping(
            tsm_data.get_match_unit_count(), create_mu_index_function(has_basic_capture_control)
        )


def make_tc_name(state_index: int, branch_index: int) -> str:
    index = state_index + branch_index * MAX_STATE_COUNT
    return f"tc{index}"


def init_fsm_tc_props(tsm_data: ILATsmData) -> Dict[str, str]:
    # Make init values for all Trigger conditions in FSM.
    unused_value = IlaMapper.get_unused_tc_value(tsm_data.get_match_unit_count())
    init_props = {
        make_tc_name(st, br): unused_value
        for st in range(MAX_STATE_COUNT)
        for br in range(MAX_CONDITION_COUNT_PER_STATE)
    }
    return init_props


def map_tsm_to_props(
    tsm_data: ILATsmData, has_basic_capture_control: bool, error_listener: ILATsmErrorListener
) -> Dict[str, str]:
    props = init_fsm_tc_props(tsm_data)
    mu_tc_props = map_tsm_mu_tc_to_props(tsm_data, has_basic_capture_control, error_listener)
    props.update(mu_tc_props)
    props.update(map_tsm_counters_to_props(tsm_data))
    tsm_row_actions = create_tsm_table_actions(tsm_data)
    props.update(map_tsm_actions_to_props(tsm_row_actions))
    return props


def dump_fsm_regs(regs: Dict[str, str], base: int = 2) -> Dict[str, str]:
    actions = [ILATsmAction(int(num, base)) for num in regs["fsm"].splitlines()]
    res = []
    it = iter(actions)
    for st in range(MAX_STATE_COUNT):
        res.append(f"\nInputs:if-elif-counter - State {st} Actions")
        for row in range(TABLE_ROWS_PER_STATE):
            if row == 4:
                res.append("")
            res.append(f"{row:03b}  " + next(it).dump())
    return {"fsm": "\n".join(res)}


def dump_tsm_regs(regs: Dict[str, str]) -> Dict[str, str]:
    # Just the fsm reg value is re-formatted.
    dump_regs = copy.copy(regs)
    dump_regs.update(dump_fsm_regs(dump_regs))
    return dump_regs



# Copyright (C) 2024, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
class ProgramLogStrategy:
    """
    Interface for the PDI Report.

    The client and server must implement the 4 empty functions.
    When implemented, the generate_report() method will generate
    an enhanced program log for use in the pdi_debug_util.

    This class lives in a common area to both cs_server and chipscopy.
    """

    def get_slr_count(self) -> int:
        """
        Return the number of SLRs in the device
        """
        ...

    def get_reg_hex_val(self, reg_name: str, slr_index: int) -> str:
        """
        Return the hex value of the named register.
        Supported registers: jtag_status, error_status.

        Example::
            get_reg_hex_val("jtag_status")
            0x042e810f0d
        """
        ...

    def get_reg_bitfield(self, reg_name: str, slr_index: int) -> str:
        """
        Return a string with lines representing the bitfield of the registers.
        Supported registers: jtag_status, error_status.

        Example::
            get_reg_bitfield("jtag_status")
            RESERVED_35 (Bits [35]): 0
            DONE (Bits [34]): 1
            JRDBK_ERROR (Bits [33]): 0
            JCONFIG_ERROR (Bits [32]): 0
            PMC_VERSION (Bits [31:28]): 2
            RESERVED_27_24 (Bits [27:24]): e
            ...

        """
        ...

    def get_plm_log(self, slr_index: int) -> str:
        """
        Return text of the PLM log.

        Example::
            get_plm_log(0)
            [0.016]****************************************
            [0.049]Xilinx Versal Platform Loader and Manager
            [0.083]Release 2024.1   Jun 13 2024  -  21:21:34
            [0.118]Platform Version: v2.0 PMC: v2.0, PS: v2.0
            [0.159]BOOTMODE: 0x0, MULTIBOOT: 0x0
            [0.188]****************************************
            ...
        """
        ...

    def generate_program_log(self) -> str:
        """
        Generate a string with an enhanced program log.
        This contains jtag_status, error_status, and the PLM log for each SLR.
        """
        report_lines = []
        for slr_index in range(self.get_slr_count()):
            report_lines.append("Enhanced PDI Log Version 1.0")
            if self.get_slr_count() > 1:
                # special case for multi-slr device
                report_lines.append(f"#################### SLR {slr_index} ####################")
            report_lines.append("JTAG Status")
            reg_val = self.get_reg_hex_val("jtag_status", slr_index)
            report_lines.append(f"JTAG STATUS: {reg_val}")
            reg_bits = self.get_reg_bitfield("jtag_status", slr_index)
            report_lines.append(reg_bits)
            report_lines.append("Error Status")
            reg_val = self.get_reg_hex_val("error_status", slr_index)
            report_lines.append(f"ERROR STATUS: {reg_val}")
            reg_bits = self.get_reg_bitfield("error_status", slr_index)
            report_lines.append(reg_bits)
            report_lines.append("PLM Log")
            plm_log = self.get_plm_log(slr_index)
            report_lines.append(plm_log)
        return "\n".join(str(e) for e in report_lines)



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from rich.tree import Tree
from typing import TYPE_CHECKING

from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.api.containers import QueryList
from chipscopy.api.ibert.aliases import GT_GROUP_KEY, HANDLE_NAME, IBERT_KEY, DISPLAY_NAME
from chipscopy.api.ibert.gt_group import GTGroup
from chipscopy.api.report import report_hierarchy
from chipscopy.utils import deprecated_api
from typing_extensions import Final, final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.client.ibert_core_client import IBERTCoreClient  # noqa


@final
class IBERT(DebugCore["IBERTCoreClient"]):
    """
    Main API class to use IBERT (Integrated Bit Error Ratio Tester) debug core.
    """

    def __init__(self, ibert_tcf_node):
        super(IBERT, self).__init__(CoreType.IBERT, ibert_tcf_node)
        _, core_info = self.core_tcf_node.initialize_architecture().popitem()

        self._gt_groups_discovery_complete: bool = False

        self.type: Final[str] = IBERT_KEY
        """Serial object type"""

        self.children: QueryList[GTGroup] = QueryList()
        """Children of IBERT"""

        self.name: str = core_info[DISPLAY_NAME]
        """Name of the IBERT core"""

        self.handle = core_info[HANDLE_NAME]
        """Handle from cs_server"""

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

    def __repr__(self) -> str:
        return self.name

    def __rich_tree__(self):
        root = Tree(self.name)

        for child in self.children:
            root.add(child.__rich_tree__())

        return root

    @property
    def gt_groups(self) -> QueryList[GTGroup]:
        """GT Group children of the IBERT core"""
        self.discover_gt_groups()
        return self.children.filter_by(type=GT_GROUP_KEY)

    def discover_gt_groups(self, *, include_uninstantiated: bool = False):
        """
        Discover all GT Groups that are instantiated by the design.

        Args:
            include_uninstantiated (bool): If you wish to include un-instantiated GT Groups in the discovery process,
                set this to 'True'.

        Returns:
            None
        """
        if self._gt_groups_discovery_complete:
            return

        gt_group_handles = self.core_tcf_node.discover_gt_groups(include_uninstantiated)
        if isinstance(gt_group_handles, str):
            gt_group_handles = [gt_group_handles]

        for handle in gt_group_handles:
            obj_info = self.core_tcf_node.get_obj_info(handle)
            self.children.append(GTGroup(obj_info, self, self.core_tcf_node))

        self._gt_groups_discovery_complete = True

    def reset(self):
        """
        Reset all RXs, TXs and PLLs in the GT Groups
        """
        for gt_group in self.gt_groups:
            gt_group.reset()

    @deprecated_api(release="2021.1", replacement="report_hierarchy(<Any serial object>)")
    def print_hierarchy(self):
        """
        Prints hierarchy of IBERTs child objects like the GT Group, GT, PLL etc to stdout.
        The format is similar to the output of the unix command - `tree`
        """
        report_hierarchy(self)



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import csv
from typing import Dict, List, TYPE_CHECKING
from dataclasses import dataclass
from collections import OrderedDict
from rich.tree import Tree

from chipscopy.dm import request
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.api.report import report_hierarchy
from chipscopy.api.hbm.hbmmc import HBMMC
from chipscopy.client.hbm_mc_client import HBMMCClient

if TYPE_CHECKING:
    from chipscopy.client.hbm_client import HBMClient


def listify(item):
    return item if isinstance(item, list) or isinstance(item, tuple) else [item]


@dataclass
class HBM(DebugCore["HBMClient"]):
    """
    This class contains the top level API functions to interact with the
    integrated HBM Stack level debug core(s) on Versal devices.

    """

    def __init__(self, hbm_node):
        super(HBM, self).__init__(CoreType.HBM, hbm_node)
        self.node_name = hbm_node.props["Name"]
        self.stack_index = self.node_name[-1]
        self.name = "hbm_stack" + self.stack_index
        self.hbm_node = self.core_tcf_node
        self.is_enabled = self.is_user_enabled()
        self.init_status = self.get_init_status()
        self.hbm_mcs: List[HBMMC] = []
        self._populate_mcs()

    def get_init_status(self):
        """
        Get the overall Middle Stack init status of HBM Stack

        Args:

        Returns:
            status in string

        """

        status = self.hbm_node.get_init_status()

        return status

    def get_overtemp_status(self):
        """
        Get the Over Temperature detection status of HBM Stack

        Args:

        Returns:
            status in string

        """
        key = "ms.reg_isr.cattrip"
        ot_status = None

        if self.is_enabled:
            results = self.hbm_node.refresh_property(key)
            status = results[key]

            if status:
                ot_status = "ERROR"
            else:
                ot_status = "OK"

        return ot_status

    def refresh_temperature(self, done: DoneHWCommand = None):
        """
        Refresh the temperature status of HBM Stack

        Args:
            done: Optional command callback that will be invoked when the response is received

        Returns:

        """

        # self.hbm_node.refresh_cal_status(done)

    def get_enabled_mcs_list(self):
        """
        Get a list of integers representing the enabled MC index under the HBM Stack

        Args:

        Returns:
            A list of integers representing the enabled MC index under the HBM Stack

        """
        en_list = []

        en_mask_tup = self.hbm_node.get_enabled_mcs_mask()
        en_mask = en_mask_tup[0]

        for mc_index in range(0, len(self.hbm_mcs)):
            if en_mask & 1:
                en_list.append(mc_index)
            en_mask = en_mask >> 1

        return en_list

    def display_status_tree(self):
        """
        Display the current status of HBM stack and the enabled HBMMCs under it

        Args:

        Returns:
            Printer outputs in the format of tree hierarchy

        """

        report_hierarchy(self)

    def get_temperature(self):
        """
        Refresh and get the current temperature of the HBM Stack

        Args:

        Returns:
            Current temperature value in integer of HBM Stack in Celcius

        """
        self.hbm_node.refresh_temp_status()

        key = "temp_value"
        result = self.hbm_node.get_property(key)

        return result[key]

    def get_mc(self, mc_index: int):
        """
        Get an HBM_MC object with a given MC index under the HBM Stack

        Args:

        Returns:
            HBM_MC object with a given MC index under the HBM Stack

        """
        mc_count = len(self.hbm_mcs)

        if (mc_index < 0) or (mc_index + 1 > mc_count):
            raise Exception("Invalid HBM MC index given.")

        return self.hbm_mcs[mc_index]

    def is_user_enabled(self):
        """
        Find out whether the HBM Stack core is user enabled or not after
        Versal device configuration

        Args:

        Returns:
            Status value in Bool

        """

        key = "en_status"
        result = self.hbm_node.get_property(key)

        if not result[key]:
            return False
        else:
            return True

    def _populate_mcs(self):
        if self.is_enabled:
            for node in self.hbm_node.manager.get_children(self.hbm_node):
                if HBMMCClient.is_compatible(node):
                    upgraded_node = self.hbm_node.manager.get_node(node.ctx, HBMMCClient)
                    HBM_MC = HBMMC(self, upgraded_node)
                    upgraded_node.api_client_wrapper = HBM_MC
                    self.hbm_mcs.append(HBM_MC)

    def save_properties(self, file_name=None, mc_index=-1, group: [str] = None):
        """
        Refresh, get, and save registers and properties data from the intended Stack or
        HBMMC to a file in CSV format

        Args:

            file_name: the file name or full path for the data file to be saved.
                       A default file name will be given if none is supplied.

            mc_index:  the target HBMMC index under the HBM Stack to point to
                       When this argument is not supplied, it defaults to just data in the Stack

            group:  Specific group names(s) that the users want to see within a block, can be a
                         single or a list of string values.
                         When this argument is not supplied, ALL groups within the block will be processed.

        Returns:

        """
        out_file = None

        if not file_name:
            # Provide a default for file
            file_name = "hbm2e_property_stack"
            file_name += self.stack_index

        headers = []
        headers.append("Name")
        headers.append("Value")
        data = {}
        groups = []
        if group:
            group = listify(group)
            groups.extend(group)

        if mc_index >= 0:
            # Users intend to fetch properties from HBMMC
            mc = self.get_mc(mc_index)
            data = mc.mc_node.refresh_property_group(groups)
            file_name += "_mc" + str(mc_index)
        else:
            # Intend for HBM Stack and MS
            self.hbm_node.refresh_temp_status()
            data = self.hbm_node.refresh_property_group(groups)

        file_name = os.path.splitext(file_name)[0] + ".csv"

        with open(file_name, "w", newline="") as out_file:
            data_writer = csv.DictWriter(out_file, headers)
            data_writer.writeheader()
            writer = csv.writer(out_file)
            for key, val in sorted(data.items()):
                if isinstance(val, int):
                    value = hex(val)
                    writer.writerow([key, value])
            print("\nInfo: Properties data are being formatted and saved to:", out_file.name)

    def get_property(self, property_names=None, done: DoneHWCommand = None):
        """
        Get the property value mapped to HBM Stack core

        Args:
            property_names: single string or list of string of property names
                (if None specified, all properties available will be queried and returned)

            done: Optional command callback that will be invoked when the response is received

        Returns:
            A dictionary, where
                Key = property name
                Val = property value

        """

        results = self.hbm_node.get_property(property_names, done)

        return results

    def __rich_tree__(self):
        root_display = self.name + " -- Status: "
        root_display += self.init_status
        root = Tree(root_display)

        for mc in self.hbm_mcs:
            if mc.is_enabled:
                display_name = mc.name + "("
                display_name += mc.mc_location + ") -- Status: "
                display_name += mc.cal_status
                root.add(display_name)

        return root

    def __str__(self):
        return self.name



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Dict, Literal

from chipscopy import dm
from chipscopy.api.ibert.aliases import DISPLAY_NAME
from chipscopy.client import core
from chipscopy.client.core import CoreClient
from chipscopy.dm.request import DoneCallback
from chipscopy.utils import listify
from chipscopy.utils.logger import log

DOMAIN_NAME = "client_ibert"


class IBERTCoreClient(CoreClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return_val = node.type and node.type == "ibert"
        return return_val

    def post_init(self):
        pass

    def get_child_with_name(self, name: str, done: DoneCallback):
        """
        Iterate through the children of this IBERT and return Node with matching name.
        """
        for child_ctx in self.children:
            child_node = self.manager.get_node(child_ctx)
            if "Name" in child_node.props and name == child_node.props["Name"]:
                done(result=child_node)
                return

        done()

    def get_child_node_with_name(self, name: str, children: [str]):
        """
        Iterate through the children of this IBERT and return Node with matching name.
        """
        for child_ctx in children:
            child_node = self.manager.get_node(child_ctx)
            if "Name" in child_node.props and name == child_node.props["Name"]:
                return child_node
            else:
                found_node = self.get_child_node_with_name(name, child_node.children)
                if found_node:
                    return found_node

    def get_node_with_name(self, name: str, done: DoneCallback):
        """
        Iterate through the node tree under this IBERT and return Node with matching name.
        """
        node = self.get_child_node_with_name(name, self.children)

        if node:
            done(result=node)
            return
        done()

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService("IBERT")

    def tier1_initialize(self, done: DoneCallback = None):
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initializing IBERT core skeleton")
        token = service.tier1_initialize(self.ctx, done_cb)
        return self.add_pending(token)

    def initialize(self, done: DoneCallback = None):
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initializing IBERT service")
        token = service.initialize(self.ctx, done_cb)
        return self.add_pending(token)

    def initialize_architecture(self, done: DoneCallback = None):
        service, done_cb = self.make_done(done)
        token = service.initialize_architecture(self.ctx, done_cb)
        return self.add_pending(token)

    def discover_gt_groups(self, include_uninstantiated: bool = False, done: DoneCallback = None):
        """
        Discover GT Groups that are part of this IBERT core.

        Args:
            include_uninstantiated (bool): Set this to True, if you wish to include GT Groups,
                un-instantiated by the design, in the list of discovered GT Groups.

            done: Callback function called when GT Groups have been discovered

        Returns:

        """
        service, done_cb = self.make_done(done)
        options = {"node_id": self.ctx, "include_uninstantiated": include_uninstantiated}
        token = service.discover_gt_groups(options, done_cb)
        return self.add_pending(token)

    def setup_gt_group(
        self, gt_group: str, done: DoneCallback = None, *, skip_post_ops: bool = False
    ):
        service, done_cb = self.make_done(done)
        options = {"node_id": self.ctx, "gt_group": gt_group, "skip_post_ops": skip_post_ops}
        token = service.setup_gt_group(options, done_cb)
        return self.add_pending(token)

    def get_obj_info(
        self,
        handle: str,
        done: DoneCallback = None,
        *,
        include_alias: bool = True,
        include_property: Literal["no", "all", "client_visible"] = "no",
    ):
        service, done_cb = self.make_done(done)
        options = {
            "node_id": self.ctx,
            "handle": handle,
            "include_alias": include_alias,
            "include_property": include_property,
        }
        token = service.get_obj_info(options, done_cb)
        return self.add_pending(token)

    def setup(self, done: DoneCallback = None):
        service, done_cb = self.make_done(done)
        options = {"node_id": self.ctx}
        token = service.setup(options, done_cb)
        return self.add_pending(token)

    def read(
        self,
        start_address: int,
        endpoint_display_name: str,
        size: int = 1,
        *,
        done: DoneCallback = None,
    ):
        """
        Command to do single or burst read.

        Args:
            start_address (int): The address to read from.

            endpoint_display_name (str): Display name for the endpoint the register is located in.

            size (int): **(Optional)** The number of words to read. Default = 1.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            str:
                Format - `"{Address as hex} : {Value in hex} : {Value as bin}"`

        Usage
            ::

                #
                >>> ibert.read(0x3000, "Quad_0")
                0x3000 : 0x0000_1234 : 0b0000_0000_0000_0000_0001_0010_0011_0100
                >>> ibert.read(0x3000, "Quad_0", 5)
                0x3000 : 0x0000_1234 : 0b0000_0000_0000_0000_0001_0010_0011_0100
                0x3004 : 0x0000_0000 : 0b0000_0000_0000_0000_0000_0000_0000_0000
                0x3008 : 0x0870_0000 : 0b0000_1000_0111_0000_0000_0000_0000_0000
                0x300C : 0x0000_0000 : 0b0000_0000_0000_0000_0000_0000_0000_0000
                0x3010 : 0x0000_0000 : 0b0000_0000_0000_0000_0000_0000_0000_0000

        """
        service, done_cb = self.make_done(done)
        options = {
            "node_id": self.ctx,
            "endpoint_name": endpoint_display_name,
            "start": start_address,
            "size": size,
        }
        token = service.read(options, done_cb)
        return self.add_pending(token)

    def write(
        self,
        start_address: int,
        data: int,
        endpoint_display_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Command to write a value to an address.

        Args:
            start_address (int): Address to write to.

            data (int): Value to write.

            endpoint_display_name (str): Display name for the endpoint the register is located in.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            None

        Usage
            ::

                #
                >>> ibert.write(0x1234, 0xABCD_EF12, "Quad_0")
                None

        """
        service, done_cb = self.make_done(done)
        options = {
            "node_id": self.ctx,
            "endpoint_name": endpoint_display_name,
            "start": start_address,
            "data": data,
        }
        token = service.write(options, done_cb)
        return self.add_pending(token)

    def get_layout(self, *, done: DoneCallback = None):
        """
        Get the hierarchy of objects in the IBERT node and the properties associated at each level.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict: Nested dictionaries with 2 keys at every level - "Properties" and "Children".

        """
        service, done_cb = self.make_done(done)
        options = {"node_id": self.ctx}
        token = service.get_layout(options, done_cb)
        return self.add_pending(token)

    def get_property(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Get value of the properties from the core object.

        Args:
            property_names (str or list[str]): A string with the name of the property OR
                a list of property name string(s).

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, str) or str: String or dictionary depending on how the command is invoked.
            See below for examples.

        Usage
            ::

                # String as first argument
                >>> ibert.get_property("<Property name>", "Quad_0")
                "<Property value">

                # List of length 1 as first argument
                >>> ibert.get_property(["<Property name>"], "Quad_0")
                "<Property value">

                # List of length 3 as first argument
                >>> ibert.get_property(
                >>>     [
                >>>         "<Property 1 name>",
                >>>         "<Property 1 name>",
                >>>         "<Property n name>",
                >>>     ],
                >>>     "Quad_0"
                >>> )
                {
                    "<Property 1 name>": "<Property value>",
                    "<Property 2 name>": "<Property value>",
                    "<Property n name>": "<Property value>",
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(property_names, str):
            property_names = [property_names]
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.get_property(options, done_cb)
        return self.add_pending(token)

    def get_property_group(self, groups: Union[str, List[str]], *, done: DoneCallback = None):
        """
        Get value of the properties in property group(s) from the core object.

        Args:
            groups (list[str] or str):  A string with the name of the property group OR
                a list of property group string(s).

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict: Dictionary with depth of 1 or 2, depending on how the command is invoked.
            See below for examples

        Usage
            ::

                # String as argument
                >>> ibert.get_property_group("<Property group name>")
                {
                    "<Property 1 name>": "<Property value>",
                    "<Property 2 name>": "<Property value>",
                    "<Property n name>": "<Property value>",
                }

                # List of length 1 as argument
                >>> ibert.get_property_group(["<Property group name>"])
                {
                    "<Property 1 name>": "<Property value>",
                    "<Property 2 name>": "<Property value>",
                    "<Property n name>": "<Property value>",
                }

                # List of length 3 as argument
                >>> ibert.get_property_group(
                >>>    [
                >>>         "<Property group 1 name>",
                >>>         "<Property group 2 name>",
                >>>         "<Property group n name>"
                >>>    ]
                >>> )
                {
                    "<Property group 1 name>":  {
                        "<Property 1 name>": "<Property value>",
                        "<Property 2 name>": "<Property value>",
                        "<Property n name>": "<Property value>",
                    },
                    "<Property group 2 name>": {
                        "<Property 1 name>": "<Property value>",
                        "<Property 2 name>": "<Property value>",
                        "<Property n name>": "<Property value>",
                    },
                    "<Property group n name>": {
                        "<Property 1 name>": "<Property value>",
                        "<Property 2 name>": "<Property value>",
                        "<Property n name>": "<Property value>",
                    }
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(groups, str):
            groups = [groups]
        options = {"node_id": self.ctx, "Groups": groups}
        token = service.get_property_group(options, done_cb)
        return self.add_pending(token)

    def set_property(
        self,
        property_dict: Dict[str, Union[int, str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Set a new value for the property in the core object.

        Args:
            property_dict (dict): A dictionary with key, value pair as the property name and
                new value for property.

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            None

        Usage
            ::

                #
                >>> ibert.set_property(
                >>>     {
                >>>         "<Property 1>": "<New value>",
                >>>         "<Property 2>": "<New value>",
                >>>         "<Property n>": "<New value>",
                >>>     },
                >>>     "Quad_0"
                >>> )
                None

        """
        service, done_cb = self.make_done(done)
        options = {
            "node_id": self.ctx,
            "Property Dict": property_dict,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.set_property(options, done_cb)
        return self.add_pending(token)

    def refresh_property(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Update value of properties from the core.

        Args:
            property_names (str or list(str)): A string with the name of the property OR a list
                of property name string(s).

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, str) or str: String or dictionary depending on how the command is invoked.
            See below for examples.

        Usage
            ::

                # String as first argument
                >>> ibert.refresh_property("<Property name>", "Quad_0")
                "<Refreshed value>"

                # List of length 1 as first argument
                >>> ibert.refresh_property(["<Property name>"], "Quad_0")
                "<Refreshed value>"

                # List of length 3 as first argument
                >>> ibert.refresh_property(
                >>>     [
                >>>         "<Property 1 name>",
                >>>         "<Property 2 name>",
                >>>         "<Property n name>"
                >>>     ],
                >>>     "Quad_0"
                >>> )
                {
                    "<Property 1 name>": "<Refreshed value>",
                    "<Property 2 name>": "<Refreshed value>",
                    .
                    .
                    .
                    "<Property n name>": "<Refreshed value>",
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(property_names, str):
            property_names = [property_names]
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.refresh_property(options, done_cb)
        return self.add_pending(token)

    def refresh_property_group(self, groups: Union[str, List[str]], *, done: DoneCallback = None):
        """
        Update the value of the properties in the property group(s) by reading from the core.

        Args:
            groups (str or list[str]): The group name to refresh as a string OR a list of group
                name as string(s).

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict - Dictionary with a depth of 1 or 2, depending on how the command is invoked.
            See below for examples.

        Usage
            ::

                # String as argument
                >>> ibert.refresh_property_group("<Property group name>")
                {
                    "<Property 1 name>": "<Refreshed value>",
                    "<Property 2 name>": "<Refreshed value>",
                    "<Property n name>": "<Refreshed value>",
                }

                # List of length 1 as argument
                >>> ibert.refresh_property_group(["<Property group name>"])
                {
                    "<Property 1 name>": "<Refreshed value>",
                    "<Property 2 name>": "<Refreshed value>",
                    "<Property n name>": "<Refreshed value>",
                }

                # List of length 3 as argument
                >>> ibert.refresh_property_group(
                >>>    [
                >>>         "<Property group 1 name>",
                >>>         "<Property group 2 name>",
                >>>         "<Property group n name>"
                >>>    ]
                >>> )
                {
                    "<Property group 1 name>": {
                        "<Property 1 name>": "<Refreshed value>",
                        "<Property 2 name>": "<Refreshed value>",
                        "<Property n name>": "<Refreshed value>",
                    },
                    "<Property group 2 name>": {
                        "<Property 1 name>": "<Refreshed value>",
                        "<Property 2 name>": "<Refreshed value>",
                        "<Property n name>": "<Refreshed value>",
                    },
                    "<Property group n name>": {
                        "<Property 1 name>": "<Refreshed value>",
                        "<Property 2 name>": "<Refreshed value>",
                        "<Property n name>": "<Refreshed value>",
                    }
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(groups, str):
            groups = [groups]
        options = {"node_id": self.ctx, "Groups": groups}
        token = service.refresh_property_group(options, done_cb)
        return self.add_pending(token)

    def commit_property(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Commit the current value of the properties in the object to the core.

        Args:
            property_names (str or list[str]): The property name to commit as a string OR a list
                of property name string(s).

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            None

        Usage
            ::

                #
                >>> ibert.commit_property("<Property name>", "Quad_0")
                None

                >>> ibert.commit_property(
                >>>     [
                >>>         "<Property 1 name">,
                >>>         "<Property 2 name">,
                >>>         "<Property n name">,
                >>>     ],
                >>>     "Quad_0"
                >>> )
                None

        """
        service, done_cb = self.make_done(done)
        if isinstance(property_names, str):
            property_names = [property_names]
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.commit_property(options, done_cb)
        return self.add_pending(token)

    def list_property_groups(self, *, done: DoneCallback = None):
        """
        Lists the available property groups.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            list(str): A list of group names

        Usage
            ::

                #
                >>> ibert.list_property_groups
                ["<Property group 1 name>", "<Property group 2 name>", "<Property group n name>"]

        """
        service, done_cb = self.make_done(done)
        options = {"node_id": self.ctx}
        token = service.list_property_groups(options, done_cb)
        return self.add_pending(token)

    def add_to_property_watchlist(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        service, done_cb = self.make_done(done)
        property_names = listify(property_names)
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.add_to_property_watchlist(options, done_cb)
        return self.add_pending(token)

    def remove_from_property_watchlist(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        service, done_cb = self.make_done(done)
        property_names = listify(property_names)
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.remove_from_property_watchlist(options, done_cb)
        return self.add_pending(token)

    def report_property(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Get detailed report about a property/properties.

        Args:
            property_names (str or list(str)): The name of the property to get report for a list
                of property name string(s).

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict: The report for a property may vary based on the type of properties

        Usage
            ::

                # Report property on a regular property
                >>> ibert.report_property("Property A", "Quad_0")
                {
                    "Property A": {
                        "Name": "Property A",
                        "Description: "This is a regular property, that can have any value",
                        "Permission": [
                            "GET",
                            "SET",
                            "COMMIT",
                            "REFRESH"
                        ],
                        "Default value": "0",
                        "Groups": [
                            "Group 1",
                            "Group 2"
                        ],
                        "Current value": "23",
                        "Value type": str
                    }
                }

                # Property B is an enumerated property
                # In this case, an additional key "Valid values" is included in the report
                >>> ibert.report_property("Property B", "Quad_0")
                {
                    "Property B":  {
                        "Name": "Property B",
                        "Description: "This is an enumerated property",
                        "Permission": [
                            "GET",
                            "SET",
                            "COMMIT",
                            "REFRESH"
                        ],
                        "Default value": "0",
                        "Groups": [
                            "Group 1",
                            "Group 2"
                        ],
                        "Current value": "23",
                        "Valid values": [
                            "0",
                            "10",
                            "14",
                            "23",
                            "50"
                        ],
                        "Value type": str
                    }
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(property_names, str):
            property_names = [property_names]
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.report_property(options, done_cb)
        return self.add_pending(token)

    def start_eye_scan(
        self,
        rx_name: str,
        scan_parameters=None,
        *,
        done: DoneCallback = None,
    ):
        service, done_cb = self.make_done(done)

        if scan_parameters is None:
            scan_parameters = dict()

        options = {"node_id": self.ctx, "RX Name": rx_name, "Scan Parameters": scan_parameters}

        token = service.start_eye_scan(options, done_cb)
        return self.add_pending(token)

    def terminate_eye_scan(self, rx_name: str, *, done: DoneCallback = None):
        service, done_cb = self.make_done(done)

        options = {"node_id": self.ctx, "RX Name": rx_name}

        token = service.terminate_eye_scan(options, done_cb)
        return self.add_pending(token)

    def get_eye_scan_parameters(self, rx_name: str = None, *, done: DoneCallback = None):
        service, done_cb = self.make_done(done)

        if rx_name is None:
            rx_name = list()

        if isinstance(rx_name, str):
            rx_name = [rx_name]

        options = {"node_id": self.ctx, "RX Name": rx_name}
        token = service.get_eye_scan_parameters(options, done_cb)
        return self.add_pending(token)

    def start_yk_scan(self, rx_name: str, *, done: DoneCallback = None):
        service, done_cb = self.make_done(done)

        options = {"node_id": self.ctx, "RX Name": rx_name}

        token = service.start_yk_scan(options, done_cb)
        return self.add_pending(token)

    def terminate_yk_scan(self, rx_name: str, *, done: DoneCallback = None):
        service, done_cb = self.make_done(done)

        options = {"node_id": self.ctx, "RX Name": rx_name}

        token = service.terminate_yk_scan(options, done_cb)
        return self.add_pending(token)



# Copyright (C) 2024, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, Callable

from chipscopy.client.jtagdevice import JtagRegister
from chipscopy.shared.ehnanced_program_log import ProgramLogStrategy


class ClientProgramLogStrategy(ProgramLogStrategy):
    """
    ChipScoPy client side implement of the Program report strategy.
    This looks up registers and log in the chipscopy client.

    This class lives on the chipscopy client side.
    """

    def __init__(self, regs: Dict[str, JtagRegister], plm_log_fn: Callable[[int], str]):
        self._slr_count = regs.get("reg.slr_count", 1)
        self._regs = {
            "jtag_status": regs.get("jtag_status"),
            "error_status": regs.get("error_status"),
        }
        self._plm_log_fn = plm_log_fn

    def get_slr_count(self) -> int:
        return self._slr_count

    def get_reg_hex_val(self, reg_name: str, slr_index: int) -> str:
        try:
            data = self._regs[reg_name].data[slr_index]
            return "0x" + data[::-1].hex()
        except KeyError:
            return f"Register {reg_name} not found"
        except IndexError:
            return f"SLR index {slr_index} out of range"

    def get_reg_bitfield(self, reg_name: str, slr_index: int) -> str:
        try:
            report_lines = []
            reg = self._regs[reg_name]
            fields: Dict = reg.fields[slr_index]
            for _, val in fields.items():
                name = val.get("name", "")
                bit_range = val.get("bit_range", "")
                int_val: int = val.get("value", 0)
                hex_value = format(int_val, "x")
                report_lines.append(f"{name} (Bits [{bit_range}]): {hex_value}")
            return "\n".join(str(e) for e in report_lines)
        except KeyError:
            return f"Register {reg_name} not found"
        except IndexError:
            return f"SLR index {slr_index} out of range"

    def get_plm_log(self, slr_index: int) -> str:
        try:
            return self._plm_log_fn(slr_index)
        except Exception as e:
            return f"Error retrieving PLM log: {str(e)}"



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2024, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytcf
from .. import protocol
from .. import services
from ..peer import Peer
from .. import _parse_params
from ..channel.AbstractChannel import AbstractChannel, ChannelListener
from ..channel import STATE_OPEN

_listeners_active = False


def _channel_closed(ci: pytcf.ChannelInfo):
    if hasattr(ci, "channel_listeners"):
        for listener in ci.channel_listeners:
            listener.onChannelClosed(None)


class ChannelPyTcf(AbstractChannel, pytcf.ChannelInfo):
    def __init__(self, chan: pytcf.ChannelInfo = None):
        pytcf.ChannelInfo.__init__(self, chan)
        self.channel_listeners = []
        self.remote_service_by_class = {}
        self.local_service_by_name = {}
        self.remote_service_by_name = {}
        self.event_listeners = {}
        self.linked_channels = set()
        self.remote_peer = Peer(_parse_params(self.url))
        self.local_peer = Peer({})
        self.state = STATE_OPEN
        services.onChannelOpened(self, self.services, self.remote_service_by_name)
        self.__makeServiceByClassMap(self.remote_service_by_name, self.remote_service_by_class)

    # @property
    # def is_closed(self):
    #     return self.ci.is_closed

    def add_linked_channel(self, chan):
        self.linked_channels.add(chan)

    def sendCommand(self, service, name, args, listener):
        s = self.remote_service_by_name.get(service)

        def done(token, error, args):
            listener.done(error, args)

        listener.token = s.send_command(name, args, done)

    def sendResult(self, token, args):
        if len(args) > 1:
            pytcf.write_response_args(token, self, args[0], args[1:])

    def sendProgress(self, token, args):
        pytcf.write_progress_args(token, self, args)

    def addChannelListener(self, listener: ChannelListener):
        global _listeners_active
        assert protocol.isDispatchThread()
        assert listener
        if not _listeners_active:
            pytcf.add_channel_close_listener(_channel_closed)
            _listeners_active = True
        self.channel_listeners.append(listener)

    def removeChannelListener(self, listener):
        assert protocol.isDispatchThread()
        self.channel_listeners.remove(listener)

    def _add_event_listener(self, service, name, handler_name):
        def event(chan):
            args = pytcf.read_command_args(chan)
            for listener in self.event_listeners[service]:
                handler = getattr(listener, handler_name)
                if handler:
                    handler(args)
        pytcf.add_event_handler(self, service, name, event)

    def addEventListener(self, service, listener):
        assert protocol.isDispatchThread()
        svc_name = str(service)
        listener.svc_name = svc_name
        lst = self.event_listeners.get(svc_name)
        if lst is None:
            lst = []
            if hasattr(listener, "_event_handlers"):
                for k, v in listener._event_handlers.items():
                    self._add_event_listener(svc_name, k, v.__name__)
        lst.append(listener)
        self.event_listeners[svc_name] = lst

    def removeEventListener(self, service, listener):
        assert protocol.isDispatchThread()
        svc_name = str(service)
        lst = self.event_listeners.get(svc_name)
        if not lst:
            return
        for i in range(len(lst)):
            if lst[i] is listener:
                del lst[i]
                return

    def close(self):
        pytcf.disconnect_channel(self)

    def terminate(self, error):
        pytcf.disconnect_channel(self)

    def getLocalPeer(self):
        assert protocol.isDispatchThread()
        return ""

    def get_local_services(self):
        assert protocol.isDispatchThread()
        return list()

    def getRemoteServices(self):
        assert protocol.isDispatchThread()
        return self.services

    def read(self):
        pass

    def writeByte(self, n):
        pass

    def flush(self):
        pass

    def stop(self):
        pass

    def __makeServiceByClassMap(self, by_name, by_class):
        for service in list(by_name.values()):
            by_class[service.__class__] = service
            for clazz in service.__class__.__bases__:
                if clazz == services.Service:
                    continue
                by_class[clazz] = service

pytcf.set_channel_class(ChannelPyTcf)



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import sys
import threading
import traceback
import socket
import time
from typing import Optional, Dict, Any, List, Union, Callable, Set, Tuple

from chipscopy.api import DMNodeListener
from chipscopy.client.jtagdevice import JtagDevice, JtagCable
from chipscopy.dm import chipscope, Node
from chipscopy.utils.logger import log
from chipscopy.utils.version import version_consistency_check
from chipscopy.client import connect as client_connect
from chipscopy.client import disconnect as client_disconnect
from chipscopy.client.util import connect_hw, process_param_str, parse_params
from chipscopy.client.view_info import ViewInfo
from chipscopy.client.server_info import ServerInfo
from chipscopy.api.containers import QueryList
from chipscopy.api.device.device import Device, FeatureNotAvailableError, DeviceState
from chipscopy.api.device.device_util import get_jtag_view_dict
from chipscopy.api.memory import Memory
from chipscopy.api.cable import Cable, discover_devices, wait_for_all_cables_ready, discover_cables

DOMAIN_NAME = "client"
log.register_domain(DOMAIN_NAME)


class Session:
    """Top level object that tracks a connection to a hardware server and optionally, chipscope
    server. To create and destroy a session, use the factory function
    create_session() and delete_session().
    """

    _session_lock = threading.RLock()

    # Global session tracking here. We keep a list of who is connected at any time.
    # Format:
    #    dict[hex(id(session))] = (hw_server, cs_server, session)
    _connected_session_dict: Dict[str, Tuple[ServerInfo, ServerInfo, "Session"]] = {}

    # cs_server channel reference counting.
    # cs_server instances are shared for the same host/port.
    # Format:
    #    dict[(normalized_host,port)] = (count, cs_server)
    _cs_server_channel_ref_count: Dict[Tuple[str, int], Tuple[int, ServerInfo]] = {}

    def __init__(
        self,
        *,
        hw_server_url: str,
        cs_server_url: Optional[str] = None,
        xvc_mm_server_url: Optional[str] = None,
        disable_core_scan: bool,
        bypass_version_check: bool,
        cable_timeout: int,
        disable_cache: bool,
        initial_device_scan: bool,
        pre_device_scan_delay: int,
        cs_server_sharing: bool,
    ):
        self._cs_server_sharing: bool = cs_server_sharing
        self._disable_core_scan: bool = disable_core_scan
        self._bypass_version_check: bool = bypass_version_check
        self._hw_server_url: str = hw_server_url
        self._cs_server_url: str = cs_server_url
        self._xvc_mm_server_url: str = xvc_mm_server_url
        self._cable_timeout = cable_timeout
        self._cables_are_initialized = False
        self._disable_cache = disable_cache
        self._register_node_listeners = not self._disable_cache
        self._initial_device_scan = initial_device_scan
        self._pre_device_scan_delay = pre_device_scan_delay

        self.hw_server: Optional[ServerInfo] = None
        self.cs_server: Optional[ServerInfo] = None

        self._need_to_scan_devices = True
        self._devices: QueryList[Device] = QueryList()
        self._device_lock = threading.Lock()

        self._dm_node_listener = DMNodeListener(self.node_callback)

    def __str__(self):
        return f"{self.handle}"

    def __repr__(self):
        return self.to_json()

    def __getitem__(self, item):
        props = self.to_dict()
        if item in props:
            return props[item]
        else:
            raise AttributeError(f"No property {str(item)}")

    def to_json(self):
        json_dict = json.dumps(self.to_dict(), indent=4)
        return json_dict

    def to_dict(self):
        ret_dict = {
            "name": self.handle,
            "hw_server_url": self._hw_server_url,
            "cs_server_url": self._cs_server_url,
            "xvc_mm_server_url": self._xvc_mm_server_url,
            "bypass_version_check": self._bypass_version_check,
            "disable_core_scan": self._disable_core_scan,
            "cable_timeout": self._cable_timeout,
            "handle": self.handle,
        }
        return ret_dict

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            traceback.print_exception(exc_type, exc_val, exc_tb)
        delete_session(self)

    def node_callback(self, action: DMNodeListener.NodeAction, node: Node, props: Optional[Set]):
        # This callback is called when nodes are added, changed, or removed.
        if node:
            valid_device_list = QueryList()
            devices = self._get_devices_with_lock()
            for device in devices:
                # Propagate node events to devices, so they can decide what to do with the event
                device.node_callback(action, node, props)
                if device.state == DeviceState.INVALID:
                    # Not typical - could be a board power off situation
                    # Anything that causes a device to be INVALID means we need to re-scan the
                    # devices and get that one removed from the list.
                    self._need_to_scan_devices = True
                else:
                    valid_device_list.append(device)

            self._set_device_with_lock(valid_device_list)

    @classmethod
    def _add_connection(
        cls, hw_server: ServerInfo, cs_server: Optional[ServerInfo], session: "Session"
    ):
        with cls._session_lock:
            cls._connected_session_dict[hex(id(session))] = (hw_server, cs_server, session)

    @classmethod
    def _remove_connection(cls, session: "Session"):
        with cls._session_lock:
            del cls._connected_session_dict[hex(id(session))]

    @classmethod
    def _get_all_sessions(cls) -> List["Session"]:
        sessions = []
        with cls._session_lock:
            for id_key, (_, _, session) in cls._connected_session_dict.items():
                sessions.append(session)
        return sessions

    @classmethod
    def disconnect_all_sessions(cls):
        sessions_to_disconnect = []
        with cls._session_lock:
            for id_key, (_, _, session) in cls._connected_session_dict.items():
                sessions_to_disconnect.append(session)
            for session in sessions_to_disconnect:
                try:
                    session.disconnect()
                except Exception:  # pragma: no cover
                    # TODO: Look into root cause later - happens intermittently in pytest infra causing tests to fail
                    pass
            cls._connected_session_dict = {}
            cls._cs_server_channel_ref_count = {}

    @staticmethod
    def _parse_url(url: str) -> Tuple[str, int]:
        param_str = process_param_str(url)
        host_port_dict = parse_params(param_str)
        host = host_port_dict.get("Host")
        port = int(host_port_dict.get("Port"))
        return host, port

    def connect_hw_server(self):
        if not self._hw_server_url:
            # A hw_server is always required when creating a session
            raise ValueError("hw_server_url must point to a valid hw_server")
        if self.hw_server:
            raise RuntimeError("hw_server already connected")
        server = Session._connect_server("hw_server", self._hw_server_url, connect_hw)
        self.hw_server = server

    def disconnect_hw_server(self):
        if not self.hw_server:
            # silently ignore disconnect calls if no server was ever connected - no harm
            return
        client_disconnect(self.hw_server)
        self.hw_server = None

    def connect_cs_server(self):
        if not self._cs_server_url:
            # A cs_server is OPTIONAL required when creating a session - so no error
            return
        if self.cs_server:
            raise RuntimeError("cs_server already connected")

        # Due to a client limitation, we can not connect multiple times to the
        # server from the same process. This causes errors due to the manager being
        # removed on second connect.
        #
        # WORKAROUND:
        # Reference count to safely share the cs_server connection
        #
        host, port = Session._parse_url(self._cs_server_url)
        ip = socket.gethostbyname(host)

        # entry = Tuple(ref_count, cs_server)
        entry = Session._cs_server_channel_ref_count.get((ip, port))
        if entry is None or not self._cs_server_sharing:
            # This is the first session - create ref_cnt entry and server
            server = Session._connect_server("cs_server", self._cs_server_url, client_connect)
            entry = (0, server)

        # Increment reference count for every session connection to same cs_server
        server = entry[1]
        entry = (entry[0] + 1, server)
        Session._cs_server_channel_ref_count[(ip, port)] = entry

        server.connect_remote(self.hw_server.url)
        if self._xvc_mm_server_url and server:
            server.connect_xvc(self._xvc_mm_server_url, self._hw_server_url)
        self.cs_server = server

    def disconnect_cs_server(self):
        if not self.cs_server:
            # silently ignore disconnect calls if no server was ever connected - no harm
            return
        host, port = Session._parse_url(self._cs_server_url)
        ip = socket.gethostbyname(host)

        # entry = Tuple(ref_count, cs_server)
        entry: Tuple[int, ServerInfo] = Session._cs_server_channel_ref_count.get((ip, port))
        if not entry:
            # This should not be possible - no entry indicates something unbalanced happened.
            raise RuntimeError("Error during disconnect - no ref_cnt for session.cs_server")

        ref_cnt, server = entry
        ref_cnt -= 1

        server.disconnect_remote(f"TCP:{self.hw_server.url}")
        self.cs_server = None

        if ref_cnt == 0 or not self._cs_server_sharing:
            # This is the last session - delete the ref count tracking entry from dict
            del Session._cs_server_channel_ref_count[(ip, port)]
            client_disconnect(server)
        else:
            # This is not the last session - just decrement the ref count
            Session._cs_server_channel_ref_count[(ip, port)] = (ref_cnt, server)

    def connect(self):
        self.connect_hw_server()

        with Session._session_lock:
            # cs_server changes are locked because they use a reference counted
            # shared cs_server - this is delicate
            self.connect_cs_server()

        Session._add_connection(self.hw_server, self.cs_server, self)

        try:
            # Quick sanity check - throws RuntimeError on version mismatch
            version_consistency_check(self.hw_server, self.cs_server, self._bypass_version_check)

            # Add event listeners
            if self._register_node_listeners:
                if self.hw_server:
                    for view_name in ["memory", "jtag", "debugcore"]:
                        self.hw_server.get_view(view_name).add_node_listener(self._dm_node_listener)
                if self.cs_server:
                    self.cs_server.get_view("chipscope").add_node_listener(self._dm_node_listener)

            if self._pre_device_scan_delay:
                time.sleep(self._pre_device_scan_delay / 1000.0)

            if self._initial_device_scan:
                # Ensures we have a set of pre-scanned devices ready to go for first use.
                # Downside is this does take up-front time, so it is optional.
                # If not done at session creation, it will be done when devices are first accessed with session.devices
                self.scan_devices()

        except RuntimeError:
            self.disconnect()
            t, v, tb = sys.exc_info()
            raise t(v).with_traceback(tb)

    def disconnect(self):
        # Cleanup registered event listeners
        if self._register_node_listeners:
            if self.hw_server:
                for view_name in ["memory", "jtag", "debugcore"]:
                    self.hw_server.get_view(view_name).remove_node_listener(self._dm_node_listener)
            if self.cs_server:
                self.cs_server.get_view("chipscope").remove_node_listener(self._dm_node_listener)

        with Session._session_lock:
            # cs_server changes are locked because they use a reference counted
            # shared cs_server - this is delicate
            self.disconnect_cs_server()
        self.disconnect_hw_server()
        Session._remove_connection(self)

    def set_param(self, params: Dict[str, Any]):
        """Generic parameter get and set for low level chipscope server params"""
        if not isinstance(params, dict):
            message = "Please provide the params to set as a dictionary!"
            log[DOMAIN_NAME].error(message)
            raise TypeError(message)
        cs_service = self.cs_server.get_sync_service("ChipScope")
        cs_service.set_css_param(params)

    def get_param(self, params: Union[str, List[str]]) -> Dict[str, str]:
        """Generic parameter get and set for low level chipscope server params"""
        if isinstance(params, str):
            params = [params]
        cs_service = self.cs_server.get_sync_service("ChipScope")
        return cs_service.get_css_param(params).get()

    @property
    def handle(self) -> str:
        if self.cs_server and self.hw_server:
            handle_str = f"{self.cs_server.url}<->{self.hw_server.url}"
        elif self.hw_server:
            handle_str = f"{self.hw_server.url}"
        else:
            handle_str = "no_hw_server<->no_cs_server"
        return handle_str

    @property
    def chipscope_view(self) -> ViewInfo:
        view = None
        if self.cs_server:
            view = self.cs_server.get_view(chipscope)
        return view

    @property
    def jtag_devices(self) -> QueryList[JtagDevice]:
        def matcher(node, key, value):
            if getattr(node, key, None) == value:
                return True
            elif node.props.get(key) == value:
                return True
            return False

        devices: QueryList[JtagDevice] = QueryList()
        devices.set_custom_match_function(matcher)
        jtag_view_dict = get_jtag_view_dict(self.hw_server)
        view = self.hw_server.get_view("jtag")
        for cable_values in jtag_view_dict.values():
            devices_dict = cable_values.get("devices", {})
            jtag_index = 0
            for device_ctx in devices_dict.keys():
                jtag_device = view.get_node(ctx=device_ctx, cls=JtagDevice)
                # TODO: Check the correct way to get the jtag_index. I couldn't find in the props so I just add 1 each
                #       time we find a device on the cable...
                jtag_device.cable_serial = cable_values["props"].get("Serial")
                jtag_device.jtag_index = jtag_index
                jtag_index += 1
                devices.append(jtag_device)
        return devices

    @property
    def jtag_cables(self) -> QueryList[JtagCable]:
        # TODO: Transition jtag_cables -> target_cables.
        #       target_cables will return a higher level wrapper that can represent a virtual cable with virtual
        #       devices in the future. jtag_cables here only support the JtagCable node.
        def matcher(node, key, value):
            if getattr(node, key, None) == value:
                return True
            elif node.props.get(key) == value:
                return True
            return False

        jtag_cables: QueryList[JtagCable] = QueryList()
        jtag_cables.set_custom_match_function(matcher)
        jtag_view_dict = get_jtag_view_dict(self.hw_server)
        view = self.hw_server.get_view("jtag")
        for jtag_cable_ctx, cable_values in jtag_view_dict.items():
            jtag_cable = view.get_node(ctx=jtag_cable_ctx, cls=JtagCable)
            jtag_cables.append(jtag_cable)
        return jtag_cables

    @property
    def cables(self) -> QueryList[Cable]:
        """Returns a list of all cables connected to the hw_server.
        Similar to vivado get_hw_targets command. target_cables may be jtag or virtual.
        """
        retval = discover_cables(
            hw_server=self.hw_server,
            cs_server=self.cs_server,
            disable_core_scan=self._disable_core_scan,
            timeout=self._cable_timeout,
        )
        return retval

    def scan_devices(self) -> QueryList[Device]:
        if not self._cables_are_initialized:
            wait_for_all_cables_ready(self.hw_server, self._cable_timeout)
            self._cables_are_initialized = True
        devices = discover_devices(
            hw_server=self.hw_server,
            cs_server=self.cs_server,
            disable_core_scan=self._disable_core_scan,
            cable_ctx=None,
            disable_cache=self._disable_cache,
        )
        self._set_device_with_lock(devices)
        return QueryList(self._devices)

    def _get_devices_with_lock(self) -> QueryList[Device]:
        with self._device_lock:
            return QueryList(self._devices)

    def _set_device_with_lock(self, devices: QueryList) -> None:
        with self._device_lock:
            self._devices = QueryList(devices)
            self._need_to_scan_devices = False

    @property
    def devices(self) -> QueryList[Device]:
        """Returns a list of devices connected to this hw_server and cable.
        Devices may span multiple jtag chains. No ordering is guaranteed.
        Devices are cached after scanning until an event changes state and
        requires a re-scan if caching is enabled.
        """
        if self._disable_cache or self._need_to_scan_devices:
            # Force scan from hardware (slow)
            devices = self.scan_devices()
        else:
            # Use the cached device list (fast)
            devices = self._get_devices_with_lock()
        return devices

    @property
    def memory(self) -> QueryList[Memory]:
        memory_node_list = QueryList()
        # This is not fast - improve it later. We don't need to query each
        # device - but it's clear and simple now.
        for device in self.devices:
            try:
                memory_nodes = device.memory
                for node in memory_nodes:
                    memory_node_list.append(node)
            except FeatureNotAvailableError:
                pass
        return memory_node_list

    @staticmethod
    def set_log_level(level: str = None):
        """
        Set the logging level for the ChipScoPy client. This applies to all sessions.
        Default is "NONE"

        Args:
            level: The minimum level to use for the logger. Valid levels are
                "TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL", "NONE"

        """

        if level is None:
            log.disable_domain(DOMAIN_NAME)

        valid_levels = ["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL", "NONE"]
        level = level.upper()
        assert level in valid_levels
        log.change_log_level(level)
        log.enable_domain(DOMAIN_NAME)

    @staticmethod
    def _connect_server(server_name: str, server_url: str, connect_func: Callable) -> ServerInfo:
        try:
            return connect_func(server_url)
        except ConnectionRefusedError:
            # Shorten stack dump, by creating new exception.
            raise ConnectionRefusedError(
                f"Connection could not be opened to {server_name} @ {server_url}.\n"
                f"  Please make sure that {server_name} is running at the URL provided, "
                f"and firewall is not blocking."
            )
        except ConnectionError:
            # Shorten stack dump, by creating new exception.
            raise ConnectionRefusedError(
                f"Connection could not be opened to {server_name} @ {server_url}.\n"
                f"  Please make sure that {server_name} is running at the URL provided, "
                f"and firewall is not blocking."
            )
        except Exception as ex:
            # Chain exceptions.
            raise Exception(
                f"Connection could not be opened to {server_name} @ {server_url}."
            ) from ex


###############################################################################
# Factory methods below
###############################################################################


def create_session(*, hw_server_url: str, cs_server_url: Optional[str] = None, **kwargs) -> Session:
    """
    Create a new session. Connect to the specified hw_server, and optionally
    the cs_server.

    - hw_server is used for programming and Versal Memory read/write operations
    - cs_server is used for higher level debug core communication

    Example 1: Default session create ::

        my_session = create_session(hw_server_url="TCP:localhost:3121",
                                    cs_server_url="TCP:localhost:3042")

    Example 2: Disable core scanning and server version checking ::

        my_session = create_session(hw_server_url="TCP:localhost:3121",
                                    cs_server_url="TCP:localhost:3042",
                                    disable_core_scan=True,
                                    bypass_version_check=True)

    Args:
        hw_server_url: Hardware server URL. Format ``TCP:<hostname>:<port>``
        cs_server_url: ChipScope server URL. Format ``TCP:<hostname>:<port>``

    Keyword Arguments:
        disable_core_scan: Set True to completely disable core scanning during discover_and_setup_debug_cores
        bypass_version_check: Set True to change hw_server and cs_server version mismatch to warning instead of error
        xvc_mm_server_url: Url for the testing xvc memory map server - For special debug core testing use cases
        cable_timeout: Seconds before timing out when detecting devices on a jtag cable
        initial_device_scan: Do an initial device scan when opening session (lazy initialization otherwise)
        pre_device_scan_delay: Time (in milliseconds) to wait after creating connections to hw_server before scanning for device
        disable_cache: Control client caching (experimental)
        auto_connect: Automatically connect to server(s) when session is created
        cs_server_sharing: Enable reference count to share cs_server connections

    Returns:
        New session object.

    """
    disable_core_scan = kwargs.get("disable_core_scan", False)
    bypass_version_check = kwargs.get("bypass_version_check", True)
    xvc_mm_server_url = kwargs.get("xvc_mm_server_url", None)
    cable_timeout = kwargs.get("cable_timeout", 4)
    initial_device_scan = kwargs.get("initial_device_scan", True)
    pre_device_scan_delay = kwargs.get("pre_device_scan_delay", 0)
    disable_cache = kwargs.get("disable_cache", False)
    auto_connect = kwargs.get("auto_connect", True)
    cs_server_sharing = kwargs.get("cs_server_sharing", False)

    # Create session even if there already exists a session with the same cs_server and hw_server
    # It *should* be safe.
    session = Session(
        cs_server_url=cs_server_url,
        hw_server_url=hw_server_url,
        xvc_mm_server_url=xvc_mm_server_url,
        disable_core_scan=disable_core_scan,
        bypass_version_check=bypass_version_check,
        cable_timeout=cable_timeout,
        disable_cache=disable_cache,
        initial_device_scan=initial_device_scan,
        pre_device_scan_delay=pre_device_scan_delay,
        cs_server_sharing=cs_server_sharing,
    )
    if auto_connect:
        session.connect()
    return session


def delete_session(session: Session):
    """
    Delete a session. Shuts down any open server connections.

    Args:
        session: The session object to delete
    """
    session.disconnect()



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from typing import Dict, List, TYPE_CHECKING
from dataclasses import dataclass
from collections import OrderedDict

from chipscopy.dm import request
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore

if TYPE_CHECKING:
    from chipscopy.client.hbm_mc_client import HBMMCClient


@dataclass
class HBMMC(DebugCore["HBMMCClient"]):
    """
    This class contains the top level API functions to interact with the
    integrated HBMMC debug core(s) under a HBM Stack on Versal devices.

    """

    def __init__(self, hbm_parent_stack, hbmmc_node):
        from chipscopy.api.hbm.hbm import HBM

        super(HBMMC, self).__init__("HBMMC", hbmmc_node)
        self.node_name = hbmmc_node.props["Name"]
        self.mc_index = hbmmc_node.props.get("mc_index")
        self.mc_location = hbmmc_node.props.get("mc_loc")
        self.name = "hbm_mc" + str(self.mc_index)
        self.mc_node = self.core_tcf_node
        self.is_enabled = self.is_user_enabled()
        self.cal_status = self.get_cal_status()
        self.hbm_stack: HBM = hbm_parent_stack

    def get_cal_status(self):
        """
        Get the overall calibration status of HBMMC under a Stack

        Args:

        Returns:
            status in string

        """

        key = "cal_status"
        result = self.mc_node.get_property(key)
        cal_status = result[key]

        return cal_status

    def is_user_enabled(self):
        """
        Find out whether the HBMMC core under a HBM Stack is user enabled or not
        after Versal device configuration

        Args:

        Returns:
            Status value in Bool

        """

        key = "mc_enabled"
        result = self.mc_node.get_property(key)

        if (key in result) and result[key]:
            return True
        else:
            return False

    def get_power_mode(self):
        """
        Get the current power mode status of HBMMC under a Stack

        Args:

        Returns:
            mode status in string

        """

        key = "mc.mc_stat"
        power_mode = None

        if self.is_enabled:
            results = self.mc_node.refresh_property_group(key)
            self_refresh = results["hbmmc_mc_stat.self_refresh"]
            pow_down = results["hbmmc_mc_stat.power_down"]
            active = results["hbmmc_mc_stat.active"]

            if self_refresh:
                power_mode = "Self Refresh"
            elif pow_down:
                power_mode = "Powered Down"
            elif active:
                power_mode = "Active"
            else:
                power_mode = "ERROR"

        return power_mode

    def get_property(self, property_names=None, done: DoneHWCommand = None):
        """
        Get the property value mapped to HBMMC core

        Args:
            property_names: single string or list of string of property names
                (if None specified, all properties available will be queried and returned)

            done: Optional command callback that will be invoked when the response is received

        Returns:
            A dictionary, where
                Key = property name
                Val = property value

        """

        results = self.mc_node.get_property(property_names, done)

        return results

    def __str__(self):
        return self.name
